import h5py
import xml.etree.cElementTree as ET
from flatdict import FlatDict
from openpnm.io import Dict, GenericIO


class XDMF(GenericIO):
    r"""
    The eXtensible Data Model Format combines XML descriptors with HDF5 data
    storage


    For more information visit the webiste:
    `XDMF.org <http://www.xdmf.org/index.php/Main_Page>`_
    """

    _header = '''<?xml version="1.0" ?>
                 <!DOCTYPE Xdmf SYSTEM "Xdmf.dtd" []>'''

    @classmethod
    def save(cls, network, phases=[], filename=''):
        r"""
        Saves data from the given objects into the specified file.

        Parameters
        ----------
        network : OpenPNM Network Object
            The network containing the desired data

        phases : list of OpenPNM Phase Objects (optional, default is none)
            A list of phase objects whose data are to be included

        Notes
        -----
        This method only saves the data, not any of the pore-scale models or
        other attributes.  To save an actual OpenPNM Project use the
        ``Workspace`` object.

        """
        project, network, phases = cls._parse_args(network=network,
                                                   phases=phases)
        network = network[0]

        if filename == '':
            filename = project.name
        f = h5py.File(filename+".hdf", "w")

        d = Dict.to_dict(network, phases=phases, interleave=True,
                         flatten=False, categorize_by=['element', 'data'])

        # Make HDF5 file with all datasets, and no groups
        D = FlatDict(d, delimiter='/')
        for item in D.keys():
            if 'U' in str(D[item][0].dtype):
                pass
            else:
                f.create_dataset(name='/'+item, shape=D[item].shape,
                                 dtype=D[item].dtype, data=D[item])
        # Add coordinate and connection information to top of HDF5 file
        f["coordinates"] = network["pore.coords"]
        f["connections"] = network["throat.conns"]

        # setup xdmf file
        root = create_root('Xdmf')
        domain = create_domain()
        grid = create_grid(Name="Structure", GridType="Uniform")

        # geometry coordinates
        row, col = f["coordinates"].shape
        dims = ' '.join((str(row), str(col)))
        hdf_loc = f.filename + ":coordinates"
        geo_data = create_data_item(value=hdf_loc, Dimensions=dims,
                                    Format='HDF', DataType="Float")
        geo = create_geometry(GeometryType="XYZ")
        geo.append(geo_data)

        # topolgy connections
        row, col = f["connections"].shape  # col first then row
        dims = ' '.join((str(row), str(col)))
        hdf_loc = f.filename + ":connections"
        topo_data = create_data_item(value=hdf_loc, Dimensions=dims,
                                     Format="HDF", NumberType="Int")
        topo = create_topology(TopologyType="Polyline",
                               NodesPerElement=str(2),
                               NumberOfElements=str(row))
        topo.append(topo_data)

        # Add pore and throat properties
        for item in D.keys():
            if item not in ['coordinates', 'connections']:
                attr_type = 'Scalar'
                shape = f[item].shape
                dims = ''.join([str(i) + ' ' for i in list(shape)[::-1]])
                hdf_loc = f.filename + ":" + item
                attr = create_data_item(value=hdf_loc,
                                        Dimensions=dims,
                                        Format='HDF',
                                        Precision='8',
                                        DataType='Float')
                name = item.replace('/', ' | ')
                if 'throat' in item:
                    Center = "Cell"
                else:
                    Center = "Node"
                el_attr = create_attribute(Name=name, Center=Center,
                                           AttributeType=attr_type)
                el_attr.append(attr)
                grid.append(el_attr)

        grid.append(topo)
        grid.append(geo)
        domain.append(grid)
        root.append(domain)

        with open(filename+'.xmf', 'w') as file:
            file.write(cls._header)
            file.write(ET.tostring(root).decode("utf-8"))

        # CLose the HDF5 file
        f.close()


def create_root(Name):
    return ET.Element(Name)


def create_domain():
    return ET.Element("Domain")


def create_geometry(GeometryType, **attribs):
    element = ET.Element('Geometry')
    element.attrib.update({'GeometryType': GeometryType})
    element.attrib.update(attribs)
    return element


def create_topology(TopologyType, NumberOfElements, **attribs):
    element = ET.Element('Topology')
    element.attrib.update({'TopologyType': TopologyType,
                           'NumberOfElements': NumberOfElements})
    if TopologyType in ['Polyline']:
        if 'NodesPerElement' not in attribs.keys():
            raise Exception('NodesPerElement must be specified')
    element.attrib.update(attribs)
    return element


def create_attribute(Name, **attribs):
    element = ET.Element('Attribute')
    element.attrib.update({'Name': Name})
    element.attrib.update(attribs)
    return element


def create_time(type='Single', Value=None):
    element = ET.Element('Attribute')
    if type == 'Single' and Value:
        element.attrib['Value'] = Value
    return element


def create_grid(Name, GridType, **attribs):
    element = ET.Element('Grid')
    element.attrib.update({'Name': Name,
                           'GridType': GridType,
                           'Section': None})
    element.attrib.update(attribs)
    if element.attrib['GridType'] is not 'Subset':
        if 'Section' in element.attrib.keys():
            del element.attrib['Section']
    return element


def create_data_item(value, Dimensions, **attribs):
    element = ET.Element('DataItem')
    element.attrib.update({'ItemType': "Uniform",
                           'Format': "XML",
                           'DatarType': "Float",
                           'Precision': "4",
                           'Rank': "1",
                           'Dimensions': Dimensions,
                           'Function': None})
    element.attrib.update(attribs)
    if element.attrib['Function'] is None:
        del element.attrib['Function']
    element.text = value
    return element



from xml.etree import ElementTree as ET
from flatdict import FlatDict
import numpy as np
from openpnm.network import GenericNetwork
from openpnm.io import GenericIO, Dict
from openpnm.utils import logging, Workspace
logger = logging.getLogger(__name__)
ws = Workspace()


class VTK(GenericIO):
    r"""
    The Visualization Toolkit (VTK) format defined by Kitware and used by
    Paraview

    Because OpenPNM data is unstructured, the actual output format is VTP,
    not VTK.
    """

    _TEMPLATE = '''
    <?xml version="1.0" ?>
    <VTKFile byte_order="LittleEndian" type="PolyData" version="0.1">
        <PolyData>
            <Piece NumberOfLines="0" NumberOfPoints="0">
                <Points>
                </Points>
                <Lines>
                </Lines>
                <PointData>
                </PointData>
                <CellData>
                </CellData>
            </Piece>
        </PolyData>
    </VTKFile>
    '''.strip()

    @classmethod
    def save(cls, network, phases=[], filename='', delim=' | ', fill_nans=None):
        r"""
        Save network and phase data to a single vtp file for visualizing in
        Paraview

        Parameters
        ----------
        network : OpenPNM Network Object
            The Network containing the data to be written

        phases : list, optional
            A list containing OpenPNM Phase object(s) containing data to be
            written

        filename : string, optional
            Filename to write data.  If no name is given the file is named
            after the network

        delim : string
            Specify which character is used to delimit the data names.  The
            default is ' | ' which creates a nice clean output in the Paraview
            pipeline viewer (e.g. net | property | pore | diameter)

        fill_nans : scalar
            The value to use to replace NaNs with.  The VTK file format does
            not work with NaNs, so they must be dealt with.  The default is
            `None` which means property arrays with NaNs are not written to the
            file.  Other useful options might be 0 or -1, but the user must
            be aware that these are not real values, only place holders.

        """
        project, network, phases = cls._parse_args(network=network,
                                                   phases=phases)

        am = Dict.to_dict(network=network, phases=phases, interleave=True,
                          categorize_by=['object', 'data'])
        am = FlatDict(am, delimiter=delim)
        key_list = list(sorted(am.keys()))

        network = network[0]
        points = network['pore.coords']
        pairs = network['throat.conns']
        num_points = np.shape(points)[0]
        num_throats = np.shape(pairs)[0]

        root = ET.fromstring(VTK._TEMPLATE)
        piece_node = root.find('PolyData').find('Piece')
        piece_node.set("NumberOfPoints", str(num_points))
        piece_node.set("NumberOfLines", str(num_throats))
        points_node = piece_node.find('Points')
        coords = VTK._array_to_element("coords", points.T.ravel('F'), n=3)
        points_node.append(coords)
        lines_node = piece_node.find('Lines')
        connectivity = VTK._array_to_element("connectivity", pairs)
        lines_node.append(connectivity)
        offsets = VTK._array_to_element("offsets", 2*np.arange(len(pairs))+2)
        lines_node.append(offsets)

        point_data_node = piece_node.find('PointData')
        cell_data_node = piece_node.find('CellData')
        for key in key_list:
            array = am[key]
            if array.dtype == np.bool:
                array = array.astype(int)
            if np.any(np.isnan(array)):
                if fill_nans is None:
                    logger.warning(key + ' has nans, will not write to file')
                    continue
                else:
                    array[np.isnan(array)] = fill_nans
            element = VTK._array_to_element(key, array)
            if (array.size == num_points) and element:
                point_data_node.append(element)
            elif (array.size == num_throats) and element:
                cell_data_node.append(element)

        if filename == '':
            filename = project.name
        filename = cls._parse_filename(filename=filename, ext='vtp')

        tree = ET.ElementTree(root)
        tree.write(filename)

        with open(filename, 'r+') as f:
            string = f.read()
            string = string.replace('</DataArray>', '</DataArray>\n\t\t\t')
            f.seek(0)
            # consider adding header: '<?xml version="1.0"?>\n'+
            f.write(string)

    @classmethod
    def load(cls, filename, project=None, delim=' | '):
        r"""
        Read in pore and throat data from a saved VTK file.

        Parameters
        ----------
        filename : string (optional)
            The name of the file containing the data to import.  The formatting
            of this file is outlined below.

        project : OpenPNM Project object
            A GenericNetwork is created and added to the specified Project.
            If no Project is supplied then one will be created and returned.

        """
        net = {}

        filename = cls._parse_filename(filename, ext='vtp')
        tree = ET.parse(filename)
        piece_node = tree.find('PolyData').find('Piece')

        # Extract connectivity
        conn_element = piece_node.find('Lines').find('DataArray')
        conns = VTK._element_to_array(conn_element, 2)
        # Extract coordinates
        coord_element = piece_node.find('Points').find('DataArray')
        coords = VTK._element_to_array(coord_element, 3)

        # Extract pore data
        for item in piece_node.find('PointData').iter('DataArray'):
            key = item.get('Name')
            array = VTK._element_to_array(item)
            net[key] = array
        # Extract throat data
        for item in piece_node.find('CellData').iter('DataArray'):
            key = item.get('Name')
            array = VTK._element_to_array(item)
            net[key] = array

        if project is None:
            project = ws.new_project()
        project = Dict.from_dict(dct=net, project=project, delim=delim)

        # Clean up data values, if necessary, like convert array's of
        # 1's and 0's into boolean.
        project = cls._convert_data(project)

        # Add coords and conns to network
        network = project.network
        network.update({'throat.conns': conns})
        network.update({'pore.coords': coords})

        return project

    @classmethod
    def _array_to_element(cls, name, array, n=1):
        dtype_map = {
            'int8': 'Int8',
            'int16': 'Int16',
            'int32': 'Int32',
            'int64': 'Int64',
            'uint8': 'UInt8',
            'uint16': 'UInt16',
            'uint32': 'UInt32',
            'uint64': 'UInt64',
            'float32': 'Float32',
            'float64': 'Float64',
            'str': 'String',
        }
        element = None
        if str(array.dtype) in dtype_map.keys():
            element = ET.Element('DataArray')
            element.set("Name", name)
            element.set("NumberOfComponents", str(n))
            element.set("type", dtype_map[str(array.dtype)])
            element.text = '\t'.join(map(str, array.ravel()))
        return element

    @classmethod
    def _element_to_array(cls, element, n=1):
        string = element.text
        dtype = element.get("type")
        array = np.fromstring(string, sep='\t')
        array = array.astype(dtype)
        if n is not 1:
            array = array.reshape(array.size//n, n)
        return array



import inspect
import scipy as _sp
import time as _time
from collections import OrderedDict


class PrintableList(list):
    r"""
    Simple subclass of ``list`` that has nice printing.  Only works flat lists.

    Example
    -------
    >>> from openpnm.utils import PrintableList
    >>> temp = ['item1', 'item2', 'item3']
    >>> print(PrintableList(temp))
    ――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――
    1     : item1
    2     : item2
    3     : item3
    ――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――

    Each line contains the result of ``print(item)`` on each item in the list

    """

    def __str__(self):
        horizontal_rule = '―' * 78
        lines = [horizontal_rule]
        self.sort()
        for i, item in enumerate(self):
            lines.append('{0:<5s} : {1}'.format(str(i + 1), item))
        lines.append(horizontal_rule)
        return '\n'.join(lines)

    def __repr__(self):
        self.sort()
        return super().__repr__()


class PrintableDict(OrderedDict):
    r"""
    Simple subclass of ``dict`` that has nicer printing.

    Example
    -------
    >>> from openpnm.utils import PrintableDict
    >>> from numpy import array as arr
    >>> d = {'item1': 1, 'item2': '1', 'item3': [1, 1], 'item4': arr([1, 1])}
    >>> print(PrintableDict(d))
    ――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――
    key                                 value
    ――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――
    item1                               1
    item2                               1
    item3                               [1, 1]
    item4                               (2,)
    ――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――

    If the item is a Numpy array the value column will contain the items'
    shape, otherwise it will contain the result of ``print(item)``

    """
    def __init__(self, *args, **kwargs):
        self._value = 'value'
        self._key = 'key'
        super().__init__(*args, **kwargs)

    def __repr__(self):
        text = dict(self).__str__()
        return text

    def __str__(self):
        header = '―' * 78
        lines = [header]
        lines.append('{0:<35s} {1}'.format(self._key, self._value))
        lines.append(header)
        for item in list(self.keys()):
            if type(self[item]) == _sp.ndarray:
                lines.append('{0:<35s} {1}'.format(item, _sp.shape(self[item])))
            else:
                lines.append('{0:<35s} {1}'.format(item, self[item]))
        lines.append(header)
        return '\n'.join(lines)


class SettingsDict(PrintableDict):
    r"""
    The SettingsDict implements the __missing__ magic method, which returns
    None instead of KeyError.  This is useful for checking the value of a
    settings without first ensuring it exists.

    Examples
    --------
    >>> from openpnm.utils import SettingsDict
    >>> sd = SettingsDict()
    >>> sd['test'] = True
    >>> print(sd['test'])
    True
    >>> print(sd['not_a_valid_key'])
    None

    """
    def __missing__(self, key):
        self[key] = None
        return self[key]


class NestedDict(dict):

    def __init__(self, mapping={}, delimiter='/'):
        super().__init__()
        self.delimiter = delimiter
        self.update(mapping)
        self.unravel()

    def __setitem__(self, key, value):
        path = key.split(self.delimiter, 1)
        if len(path) > 1:
            if path[0] not in self.keys():
                self[path[0]] = NestedDict(delimiter=self.delimiter)
            self[path[0]][path[1]] = value
        else:
            super().__setitem__(key, value)

    def __missing__(self, key):
        self[key] = NestedDict(delimiter=self.delimiter)
        return self[key]

    def unravel(self):
        for item in self.keys():
            self[item] = self.pop(item)

    def to_dict(self, dct=None):
        if dct is None:
            dct = self
        plain_dict = dict()
        for key in dct.keys():
            value = dct[key]
            if hasattr(value, 'keys'):
                plain_dict[key] = self.to_dict(value)
            else:
                plain_dict[key] = value
        return plain_dict

    def keys(self, dicts=True, values=True):
        k = list(super().keys())
        new_keys = []
        for item in k:
            if hasattr(self[item], 'keys'):
                if dicts:
                    new_keys.append(item)
            else:
                if values:
                    new_keys.append(item)
        return new_keys

    def __str__(self):
        def print_level(self, p='', indent='-'):
            for item in self.keys():
                if hasattr(self[item], 'keys'):
                    p = print_level(self[item], p=p, indent=indent + indent[0])
                elif indent[-1] != ' ':
                    indent = indent + ''
                p = indent + item + '\n' + p
            return(p)
        p = print_level(self)
        return p


class HealthDict(PrintableDict):
    r"""
    This class adds a 'health' check to a standard dictionary.  This check
    looks into the dict values, and considers empty lists as healthy and all
    else as unhealthy.  If one or more entries is 'unhealthy' the health method
    returns False.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _get_health(self):
        health = True
        for item in list(self.keys()):
            if self[item] != []:
                health = False
        return health

    health = property(fget=_get_health)


def tic():
    r"""
    Homemade version of matlab tic and toc function, tic starts or resets
    the clock, toc reports the time since the last call of tic.

    See Also
    --------
    toc

    """
    global _startTime_for_tictoc
    _startTime_for_tictoc = _time.time()


def toc(quiet=False):
    r"""
    Homemade version of matlab tic and toc function, tic starts or resets
    the clock, toc reports the time since the last call of tic.

    Parameters
    ----------
    quiet : Boolean
        If False (default) then a message is output to the console.  If True
        the message is not displayed and the elapsed time is returned.

    See Also
    --------
    tic

    """
    if '_startTime_for_tictoc' in globals():
        t = _time.time() - _startTime_for_tictoc
        if quiet is False:
            print('Elapsed time in seconds: ', t)
        else:
            return t
    else:
        raise Exception('Start time not set, call tic first')


def unique_list(input_list):
    r"""
    For a given list (of points) remove any duplicates
    """
    output_list = []
    if len(input_list) > 0:
        dim = _sp.shape(input_list)[1]
        for i in input_list:
            match = False
            for j in output_list:
                if dim == 3:
                    if i[0] == j[0] and i[1] == j[1] and i[2] == j[2]:
                        match = True
                elif dim == 2:
                    if i[0] == j[0] and i[1] == j[1]:
                        match = True
                elif dim == 1:
                    if i[0] == j[0]:
                        match = True
            if match is False:
                output_list.append(i)
    return output_list


def flat_list(input_list):
    r"""
    Given a list of nested lists of arbitrary depth, returns a single level or
    'flat' list.

    """
    x = input_list
    if isinstance(x, list):
        return [a for i in x for a in flat_list(i)]
    else:
        return [x]


def sanitize_dict(input_dict):
    r"""
    Given a nested dictionary, ensures that all nested dicts are normal
    Python dicts.  This is necessary for pickling, or just converting
    an 'auto-vivifying' dict to something that acts normal.
    """
    plain_dict = dict()
    for key in input_dict.keys():
        value = input_dict[key]
        if hasattr(value, 'keys'):
            plain_dict[key] = sanitize_dict(value)
        else:
            plain_dict[key] = value
    return plain_dict


def methods_to_table(obj):
    r"""
    """
    parent = obj.__class__.__mro__[1]
    temp = inspect.getmembers(parent, predicate=inspect.isroutine)
    parent_funcs = [i[0] for i in temp if not i[0].startswith('_')]

    temp = inspect.getmembers(obj.__class__, predicate=inspect.isroutine)
    obj_funcs = [i[0] for i in temp if not i[0].startswith('_')]
    funcs = set(obj_funcs).difference(set(parent_funcs))

    row = '+' + '-'*22 + '+' + '-'*49 + '+'
    fmt = '{0:1s} {1:20s} {2:1s} {3:47s} {4:1s}'
    lines = []
    lines.append(row)
    lines.append(fmt.format('|', 'Method', '|', 'Description', '|'))
    lines.append(row.replace('-', '='))
    for i, item in enumerate(funcs):
        try:
            s = getattr(obj, item).__doc__.strip()
            end = s.find('\n')
            if end > 47:
                s = s[:44] + '...'
            lines.append(fmt.format('|', item, '|', s[:end], '|'))
            lines.append(row)
        except AttributeError:
            pass
    return '\n'.join(lines)


def models_to_table(obj, params=True):
    r"""
    Converts a ModelsDict object to a ReST compatible table

    Parameters
    ----------
    obj : OpenPNM object
        Any object that has a ``models`` attribute

    params : boolean
        Indicates whether or not to include a list of parameter
        values in the table.  Set to False for just a list of models, and
        True for a more verbose table with all parameter values.
    """
    if not hasattr(obj, 'models'):
        raise Exception('Received object does not have any models')
    row = '+' + '-'*4 + '+' + '-'*22 + '+' + '-'*18 + '+' + '-'*26 + '+'
    fmt = '{0:1s} {1:2s} {2:1s} {3:20s} {4:1s} {5:16s} {6:1s} {7:24s} {8:1s}'
    lines = []
    lines.append(row)
    lines.append(fmt.format('|', '#', '|', 'Property Name', '|', 'Parameter',
                            '|', 'Value', '|'))
    lines.append(row.replace('-', '='))
    for i, item in enumerate(obj.models.keys()):
        prop = item
        if len(prop) > 20:
            prop = item[:17] + "..."
        temp = obj.models[item].copy()
        model = str(temp.pop('model')).split(' ')[1]
        lines.append(fmt.format('|', str(i+1), '|', prop, '|', 'model:',
                                '|', model, '|'))
        lines.append(row)
        if params:
            for param in temp.keys():
                p1 = param
                if len(p1) > 16:
                    p1 = p1[:14] + '...'
                p2 = str(temp[param])
                if len(p2) > 24:
                    p2 = p2[:21] + '...'
                lines.append(fmt.format('|', '', '|', '', '|', p1, '|',
                                        p2, '|'))
                lines.append(row)
    return '\n'.join(lines)


def conduit_lengths(network, throats=None, mode='pore'):
    r"""
    Return the respective lengths of the conduit components defined by the throat
    conns P1 T P2
    mode = 'pore' - uses pore coordinates
    mode = 'centroid' uses pore and throat centroids
    """
    if throats is None:
        throats = network.throats()
    Ps = network['throat.conns']
    pdia = network['pore.diameter']

    if mode == 'centroid':
        try:
            pcentroids = network['pore.centroid']
            tcentroids = network['throat.centroid']
            if _sp.sum(_sp.isnan(pcentroids)) + _sp.sum(_sp.isnan(tcentroids)) > 0:
                mode = 'pore'
            else:
                plen1 = _sp.sqrt(_sp.sum(_sp.square(pcentroids[Ps[:, 0]] -
                                         tcentroids), 1))-network['throat.length']/2
                plen2 = _sp.sqrt(_sp.sum(_sp.square(pcentroids[Ps[:, 1]] -
                                         tcentroids), 1))-network['throat.length']/2
        except KeyError:
            mode = 'pore'
    if mode == 'pore':
        # Find half-lengths of each pore
        pcoords = network['pore.coords']
        # Find the pore-to-pore distance, minus the throat length
        lengths = _sp.sqrt(_sp.sum(_sp.square(pcoords[Ps[:, 0]] -
                                   pcoords[Ps[:, 1]]), 1)) - network['throat.length']
        lengths[lengths < 0.0] = 2e-9
        # Calculate the fraction of that distance from the first pore
        try:
            fractions = pdia[Ps[:, 0]]/(pdia[Ps[:, 0]] + pdia[Ps[:, 1]])
            # Don't allow zero lengths
#            fractions[fractions == 0.0] = 0.5
#            fractions[fractions == 1.0] = 0.5
        except:
            fractions = 0.5
        plen1 = lengths*fractions
        plen2 = lengths*(1-fractions)

    return _sp.vstack((plen1, network['throat.length'], plen2)).T[throats]



from openpnm.core import Base
import numpy as np


class Subdomain(Base):
    r"""
    This subclass of the Base class provides the ability assign the object
    to specific locations (pores and throats) in the domain.  This class
    is subclassed by GenericGeometry and GenericPhysics.

    Notes
    -----
    The following table list the two methods added to Base by this subclass.

    +---------------------+---------------------------------------------------+
    | Methods             | Description                                       |
    +=====================+===================================================+
    | ``add_locations``   | Specified which pores and throats the object      |
    |                     | should be assigned to                             |
    +---------------------+---------------------------------------------------+
    | ``drop_locations``  | Removes the object from the specified pores and   |
    |                     | throats                                           |
    +---------------------+---------------------------------------------------+
    | ``_get_boss``       | Geomtry and Physics objects are subservient to    |
    |                     | Network and Phase objects, respectively, so this  |
    |                     | method returns a handle to the appropriate *boss* |
    +---------------------+---------------------------------------------------+

    Also listed above is a hidden method that might be useful.  The act of
    assign a Subdomain object to a subset of pores or throats basically amounts
    to creating a list in the *boss* object with the Subdomain's name, like
    ``'pore.geo_1'``, with True values where ``geo_1`` applies and False
    elsewhere.  Changing the locations of objects is just a matter of changing
    the locations of the True's and False's.

    The Project object has two methods, ``check_geometry_health`` and
    ``check_physics_health`` that look to make sure all locations are assigned
    to one and only one Geometry and/or Physics.

    """

    def __getitem__(self, key):
        element = key.split('.')[0]
        # Find boss object (either phase or network)
        if self._isa('physics'):
            boss = self.project.find_phase(self)
        else:
            boss = self.project.network
        # Get values if present, or regenerate them
        vals = self.get(key)
        # If still not found, check with boss object
        if vals is None:
            inds = boss._get_indices(element=element, labels=self.name)
            vals = boss[key][inds]
        return vals

    def __setitem__(self, key, value):
        if self.project:
            # Find boss object (either phase or network)
            if self._isa('phase'):
                boss = self.project.find_phase(self)
            else:
                boss = self.project.network
            if key in set(boss.keys()).difference({'pore.all', 'throat.all'}):
                raise Exception(key + ' already exists on ' + boss.name)
        super().__setitem__(key, value)

    def add_locations(self, pores=[], throats=[]):
        r"""
        Adds associations between an object and its boss object at the
        given pore and/or throat locations.

        Parameters
        ----------
        pores and throats : array_like
            The pore and/or throat locations for which the association should
            be added.  These indices are for the full domain.

        Notes
        -----
        For *Physics* objects, the boss is the *Phase* with which it was
        assigned, while for *Geometry* objects the boss is the *Network*.

        """
        boss = self._get_boss()
        pores = boss._parse_indices(pores)
        throats = boss._parse_indices(throats)
        if len(pores) > 0:
            self._set_locations(element='pore', indices=pores, mode='add')
        if len(throats) > 0:
            self._set_locations(element='throat', indices=throats, mode='add')

    def drop_locations(self, pores=[], throats=[], complete=False):
        r"""
        Removes association between an objectx and its boss object at the
        given pore and/or throat locations.

        Parameters
        ----------
        pores and throats : array_like
            The pore and/or throat locations from which the association should
            be removed.  These indices refer to the full domain.

        complete : boolean (default is ``False``)
            If ``True`` then *all* pore and throat associations are removed
            along with any trace that the objects were associated.

        Notes
        -----
        For *Physics* objects, the boss is the *Phase* with which it was
        assigned, while for *Geometry* objects the boss is the *Network*.

        """
        boss = self._get_boss()
        pores = boss._parse_indices(pores)
        throats = boss._parse_indices(throats)
        if complete:
            pores = boss.pores(self.name)
            throats = boss.throats(self.name)
        if len(pores) > 0:
            self._set_locations(element='pore', indices=pores, mode='drop',
                                complete=complete)
        if len(throats) > 0:
            self._set_locations(element='throat', indices=throats, mode='drop',
                                complete=complete)

    def _set_locations(self, element, indices, mode, complete=False):
        r"""
        This private method is called by ``set_locations`` and
        ``remove_locations`` as needed.

        """
        boss = self._get_boss()
        element = self._parse_element(element=element, single=True)

        # Make sure label array exists in boss
        if (element+'.'+self.name) not in boss.keys():
            boss[element+'.'+self.name] = False

        # Check to ensure indices aren't already assigned
        if mode == 'add':
            if self._isa('geometry'):
                objs = self.project.geometries().keys()
            else:
                objs = self.project.physics().keys()
            for name in objs:
                if element+'.'+name in boss.keys():
                    if np.any(boss[element+'.'+name][indices]):
                        raise Exception('Given indices are already assigned ' +
                                        'to ' + name)

        # Find mask of existing locations (network indexing)
        mask = boss[element+'.'+self.name]
        # Update mask with new locations (either add or remove)
        if mode == 'add':
            mask = mask + boss._tomask(indices=indices, element=element)
        elif mode == 'drop':
            mask = mask * (~boss._tomask(indices=indices, element=element))
        # Change size of all arrays on self
        for item in self.keys(element=element, mode='all'):
            self.update({item: boss[item][mask]})
        # Update label array in network
        boss[element+'.'+self.name] = mask
        # Remove label from boss if ALL locations are removed
        if mode == 'drop':
            if ~np.any(boss[element+'.'+self.name]):
                if complete:
                    del boss[element+'.'+self.name]
                else:
                    boss[element+'.'+self.name] = False

    def _get_boss(self):
        r"""
        Find boss object of the current Subdomain object.

        Returns
        -------
        If the current object is a Geometry, then returns the Network, if it
        is a Physics then returns the Phase it was assigned to during
        instantiation.

        """
        if self._isa('physics'):
            boss = self.project.find_phase(self)
        if self._isa('geometry'):
            boss = self.project.network
        return boss



import numpy as np
import scipy.sparse as sprs
from openpnm.algorithms import ReactiveTransport
from openpnm.utils import logging
logger = logging.getLogger(__name__)


class TransientReactiveTransport(ReactiveTransport):
    r"""
    A subclass of GenericTransport to perform transient and steady simulations.
    """

    def __init__(self, settings={}, **kwargs):
        self.settings.update({'t_initial': 0,
                              't_final': 10,
                              't_step': 0.1,
                              't_output': 1e+08,
                              't_tolerance': 1e-06,
                              'r_tolerance': 1e-04,
                              't_scheme': 'implicit'})
        super().__init__(**kwargs)
        self.settings.update(settings)
        self._A_steady = None  # Initialize the steady sys of eqs A matrix

    def setup(self, phase=None, t_initial='', t_final='', t_step='',
              t_output='', t_tolerance='', t_scheme='', **kwargs):
        if phase:
            self.settings['phase'] = phase.name
        if t_initial:
            self.settings['t_initial'] = t_initial
        if t_final:
            self.settings['t_final'] = t_final
        if t_step:
            self.settings['t_step'] = t_step
        if t_output:
            self.settings['t_output'] = t_output
        if t_tolerance:
            self.settings['t_tolerance'] = t_tolerance
        if t_scheme:
            self.settings['t_scheme'] = t_scheme
        self.settings.update(kwargs)

    def set_IC(self, values):
        r"""
        """
        self[self.settings['quantity']] = values

    def _t_update_A(self):
        r"""
        """
        network = self.project.network
        Vi = network['pore.volume']
        dt = self.settings['t_step']
        s = self.settings['t_scheme']
        if (s == 'implicit'):
            f1, f2 = 1, 1
        elif (s == 'cranknicolson'):
            f1, f2 = 0.5, 1
        elif (s == 'steady'):
            f1, f2 = 1, 0
        # Compute A (operations involve conversion to 'csr')
        A = ((f2/dt) * sprs.coo_matrix.multiply(
            sprs.coo_matrix(np.reshape(Vi, (self.Np, 1)), shape=(self.Np,)),
            sprs.identity(self.Np, format='coo')) + f1 * self._A_steady)
        # Convert A to 'coo' format to apply BCs
        A = sprs.coo_matrix(A)
        self._A = A
        return A

    def _t_update_b(self):
        r"""
        """
        network = self.project.network
        phase = self.project.phases()[self.settings['phase']]
        Vi = network['pore.volume']
        dt = self.settings['t_step']
        s = self.settings['t_scheme']
        if (s == 'implicit'):
            f1, f2, f3 = 1, 1, 0
        elif (s == 'cranknicolson'):
            f1, f2, f3 = 0.5, 1, 0
        elif (s == 'steady'):
            f1, f2, f3 = 1, 0, 1
        x_old = self[self.settings['quantity']]
        b = (f2*(1-f1)*(-self._A_steady)*x_old +
             f2*(Vi/dt)*x_old +
             f3*np.zeros(shape=(self.Np, ), dtype=float))
        self._update_physics()
        for item in self.settings['sources']:
            Ps = self.pores(item)
            # Update b
            b[Ps] = b[Ps] - f2*(1-f1)*(phase[item+'.'+'rate'][Ps])
        self._b = b
        return b

    def run(self, t=None):
        r"""
        """
        logger.info('―'*80)
        logger.info('Running TransientTransport')
        # If solver used in steady mode, no need to add ICs
        if (self.settings['t_scheme'] == 'steady'):
            self[self.settings['quantity']] = 0
        # Create a scratch b from IC
        self._b = (self[self.settings['quantity']]).copy()
        self._apply_BCs()
        # Save A matrix (with BCs applied) of the steady sys of eqs
        self._A_steady = (self._A).copy()
        # Save the initial field with the boundary conditions applied
        self[self.settings['quantity']] = (self._b).copy()
        # Override A and b according to t_scheme and apply BCs
        self._t_update_A()
        self._t_update_b()
        self._apply_BCs()
        self._A_t = (self._A).copy()
        self._b_t = (self._b).copy()
        if t is None:
            t = self.settings['t_initial']
        # Create S1 & S1 for 1st Picard's iteration
        self._update_physics()

        self._run_transient(t=t)

    def _run_transient(self, t):
        tf = self.settings['t_final']
        dt = self.settings['t_step']
        to = self.settings['t_output']
        tol = self.settings['t_tolerance']
        s = self.settings['t_scheme']
        res_t = 1  # Initialize the residual

        # Make sure 'tf' and 'to' are multiples of 'dt'
        tf = tf + (dt-(tf % dt))*((tf % dt) != 0)
        to = to + (dt-(to % dt))*((to % dt) != 0)
        self.settings['t_final'] = tf
        self.settings['t_output'] = to
        outputs = np.append(np.arange(t+to, tf, to), tf)

        if (s == 'steady'):  # If solver in steady mode, do one iteration
            logger.info('    Running in steady mode')
            x_old = self[self.settings['quantity']]
            self._t_run_reactive(x=x_old)
            x_new = self[self.settings['quantity']]

        else:  # Do time iterations
            # Export the initial field (t=t_initial)
            vals = self[self.settings['quantity']]
            self[self.settings['quantity']+'_initial'] = vals
            for time in np.arange(t+dt, tf+dt, dt):
                if (res_t >= tol):  # Check if the steady state is reached
                    logger.info('    Current time step: '+str(time)+' s')
                    x_old = self[self.settings['quantity']]

                    self._t_run_reactive(x=x_old)
                    x_new = self[self.settings['quantity']]

                    # Compute the residual
                    res_t = np.sum(np.absolute(x_old**2 - x_new**2))
                    logger.info('        Residual: '+str(res_t))
                    # Output transient solutions. Round time to ensure every
                    # value in outputs is exported.
                    if round(time, 12) in outputs:
                        ind = np.where(outputs == round(time, 12))[0][0]
                        self[self.settings['quantity']+'_'+str(ind)] = x_new
                        logger.info('        Exporting time step: ' +
                                    str(time)+' s')
                    # Update b and apply BCs
                    self._t_update_b()
                    self._apply_BCs()
                    self._b_t = (self._b).copy()
                else:  # Stop time iterations if residual < t_tolerance
                    self[self.settings['quantity'] + '_steady'] = x_new
                    logger.info('        Exporting time step: '+str(time)+' s')
                    break
            if (round(time, 12) == tf):
                logger.info('    Maximum time step reached: '+str(time)+' s')
            else:
                logger.info('    Transient solver converged after: ' +
                            str(time)+' s')

    def _t_run_reactive(self, x):
        if x is None:
            x = np.zeros(shape=[self.Np, ], dtype=float)
        self[self.settings['quantity']] = x
        relax = self.settings['relaxation_quantity']
        res = 1e+06
        for itr in range(int(self.settings['max_iter'])):
            if res >= self.settings['tolerance']:
                logger.info('Tolerance not met: ' + str(res))
                self[self.settings['quantity']] = x
                self._A = (self._A_t).copy()
                self._b = (self._b_t).copy()
                self._apply_BCs()
                self._apply_sources()
                x_new = self._solve()
                # Relaxation
                x_new = relax*x_new + (1-relax)*self[self.settings['quantity']]
                self[self.settings['quantity']] = x_new
                res = np.sum(np.absolute(x**2 - x_new**2))
                x = x_new
            elif res < self.settings['tolerance']:
                logger.info('Solution converged: ' + str(res))
                break
        return x_new



import numpy as np
from scipy.sparse.csgraph import laplacian
from openpnm.algorithms import ReactiveTransport
from openpnm.utils import logging
logger = logging.getLogger(__name__)


class AdvectionDiffusion(ReactiveTransport):
    r"""
    A subclass of GenericTransport to simulate advection diffusion.

    """

    def __init__(self, settings={}, **kwargs):
        super().__init__(**kwargs)
        # Set some default settings
        self.settings.update({'quantity': 'pore.concentration',
                              'diffusive_conductance':
                              'throat.diffusive_conductance',
                              'hydraulic_conductance':
                              'throat.hydraulic_conductance',
                              'pressure': 'pore.pressure',
                              's_scheme': 'powerlaw'})
        # Apply any received settings to overwrite defaults
        self.settings.update(settings)

    def setup(self, phase=None, quantity='', diffusive_conductance='',
              hydraulic_conductance='', pressure='', s_scheme='', **kwargs):
        r"""

        """
        if phase:
            self.settings['phase'] = phase.name
        if quantity:
            self.settings['quantity'] = quantity
        if diffusive_conductance:
            self.settings['diffusive_conductance'] = diffusive_conductance
        if hydraulic_conductance:
            self.settings['hydraulic_conductance'] = hydraulic_conductance
        if pressure:
            self.settings['pressure'] = pressure
        if s_scheme:
            self.settings['s_scheme'] = s_scheme
        super().setup(**kwargs)

    def _build_A(self, force=False):
        s_dis = self.settings['s_scheme']
        network = self.project.network
        phase = self.project.phases()[self.settings['phase']]
        conns = network['throat.conns']

        P = phase[self.settings['pressure']]
        gh = phase[self.settings['hydraulic_conductance']]
        gd = phase[self.settings['diffusive_conductance']]
        gd = np.tile(gd, 2)

        Qij = -gh*np.diff(P[conns], axis=1).squeeze()
        Qij = np.append(Qij, -Qij)

        if force:
            self._pure_A = None
        if self._pure_A is None:
            if (s_dis == 'upwind'):
                w = gd + np.maximum(0, -Qij)
                A = network.create_adjacency_matrix(weights=w)
            elif (s_dis == 'hybrid'):
                w = np.maximum(0, np.maximum(-Qij, gd-Qij/2))
                A = network.create_adjacency_matrix(weights=w)
            elif (s_dis == 'powerlaw'):
                Peij = np.absolute(Qij/gd)
                w = gd*np.maximum(0, (1-0.1*Peij)**5) + np.maximum(0, -Qij)
                A = network.create_adjacency_matrix(weights=w)
            A = laplacian(A)
            self._pure_A = A
        self.A = self._pure_A.copy()



import time
import pickle
import h5py
import numpy as np
import openpnm
from copy import deepcopy
from pathlib import Path
from openpnm.utils import SettingsDict, HealthDict, PrintableList, Workspace
ws = Workspace()


class Project(list):
    r"""
    This class provides a container for all OpenPNM objects in a given
    simulation.

    A simulation is defined as a Network and all of it's associated objects.
    When instantiating a Network, a Project can be passed as an argument, but
    if not given one is created.  When instantiating any other object either
    a Network or a Project can be supplied.  In the former case, the
    Network's Project is retrieved and used.  The end result is that all
    objects are stored in a specific Project.

    The Project to which any object belongs can be retrieved with
    ``obj.project``.  Conversely, printing a Project displays a list of all
    objects it contains.

    Moreover, all Projects are registered with the Workspace.  Since there can
    be only instance of the Workspace it is possible to view all open Projects
    by printing the Workspace.

    See Also
    --------
    Workspace

    Notes
    -----
    The following table shows all the methods that are available on the Project
    objects along with a very brief description:

    +------------------+-------------------------------------------------+
    | Method           | Description                                     |
    +==================+=================================================+
    | export_data      | Export the pore and throat data from the giv... |
    +------------------+-------------------------------------------------+
    | import_data      |                                                 |
    +------------------+-------------------------------------------------+
    | purge_object     | Remove an object from the Project.  This rem... |
    +------------------+-------------------------------------------------+
    | load_object      | Loads a single object from a file               |
    +------------------+-------------------------------------------------+
    | save_object      | Saves the given object to a file                |
    +------------------+-------------------------------------------------+
    | find_physics     | Find the Physics object(s) associated with a... |
    +------------------+-------------------------------------------------+
    | find_geometry    | Find the Geometry associated with a given Ph... |
    +------------------+-------------------------------------------------+
    | find_phase       | Find the Phase associated with a given object.  |
    +------------------+-------------------------------------------------+
    | check_geometr... | Perform a check to find pores with overlappi... |
    +------------------+-------------------------------------------------+
    | check_physics... | Perform a check to find pores which have ove... |
    +------------------+-------------------------------------------------+

    """

    def __init__(self, *args, **kwargs):
        name = kwargs.pop('name', None)
        super().__init__(*args, **kwargs)
        # Register self with workspace
        ws[name] = self
        self._grid = {}
        self.settings = SettingsDict()
        self.comments = 'Using OpenPNM ' + openpnm.__version__

    def extend(self, obj):
        r"""
        This function is used to add objects to the project.  Arguments can
        be single OpenPNM objects, an OpenPNM project list, or a plain list of
        OpenPNM objects.

        """
        if type(obj) is not list:
            obj = [obj]
        for item in obj:
            if hasattr(item, '_mro'):
                if 'GenericNetwork' in item._mro():
                    if self.network:
                        raise Exception('Project already has a network')
                # Must use append since extend breaks the dicts up into
                # separate objects, while append keeps it as a single object.
                super().append(item)
            else:
                raise Exception('Only OpenPNM objects can be added')

    def append(self, obj):
        r"""
        The Project (a list) must be kept as a flat list, so the append
        function, which can normally be used to insert a list into a list, is
        overloaded to basically prevent the normal append operation and simply
        calls ``extend``.

        """
        self.extend(obj)

    def remove(self, obj):
        r"""
        The given object is removed from the project

        This removes the object, along with all it's labels in associated
        objects, but does NOT remove the associated objects.

        See Also
        -------
        purge_object

        """
        self.purge_object(obj, deep=False)

    def pop(self, index):
        r"""
        The object at the given index is removed from the list and returned.

        Notes
        -----
        This method uses ``purge_object`` to perform the actual removal of the
        object. It is reommended to just use that directly instead.

        See Also
        --------
        purge_object

        """
        obj = self[index]
        self.purge_object(obj, deep=False)
        return obj

    def insert(self, index, obj):
        r"""
        Inserts the supplied object at the specified index in the Project list

        Notes
        -----
        The order of the objects in an OpenPNM Project lists do not matter, so
        it is recommended to just use ``append`` instead.

        See Also
        --------
        append
        extend

        """
        self.extend(obj)

    def clear(self, objtype=[]):
        r"""
        Clears objects from the project entirely or selectively, depdening on
        the received arguments.

        Parameters
        ----------
        objtype : list of strings
            A list containing the object type(s) to be removed.  If no types
            are specified, then all objects are removed.  To clear only objects
            of a specific type, use *'network'*, *'geometry'*, *'phase'*,
            *'physics'*, or *'algorithm'*.  It's also possible to use
            abbreviations, like *'geom'*.

        """
        if len(objtype) == 0:
            super().clear()
        else:
            names = [obj.name for obj in self]
            for name in names:
                try:
                    obj = self[name]
                    for t in objtype:
                        if obj._isa(t):
                            self.purge_object(obj)
                except KeyError:
                    pass

    def copy(self, name=None):
        r"""
        Creates a deep copy of the current project

        A deep copy means that new, unique versions of all the objects are
        created but with identical data and properties.

        Parameters
        ----------
        name : string
            The name to give to the new project.  If not supplied, a name
            is automatically generated.

        Returns
        -------
        A new Project object containing copies of all objects

        """
        if name is None:
            name = ws._gen_name()
        proj = deepcopy(self)
        ws[name] = proj
        return proj

    @property
    def workspace(self):
        return ws

    def _set_name(self, name):
        if name is None:
            name = ws._gen_name()
        ws[name] = self

    def _get_name(self):
        for key in ws.keys():
            if ws[key] is self:
                return key

    name = property(fget=_get_name, fset=_set_name)

    def __getitem__(self, key):
        if type(key) == str:
            obj = None
            for item in self:
                if item.name == key:
                    obj = item
            if obj is None:
                raise KeyError(key)
        else:
            obj = super().__getitem__(key)
        return obj

    def find_phase(self, obj):
        r"""
        Find the Phase associated with a given object.

        Parameters
        ----------
        obj : OpenPNM Object
            Can either be a Physics or Algorithm object

        Returns
        -------
        An OpenPNM Phase object.

        Raises
        ------
        If no Phase object can be found, then an Exception is raised.
        """
        # If received phase, just return self
        if obj._isa('phase'):
            return obj
        # If phase happens to be in settings (i.e. algorithm), look it up
        if 'phase' in obj.settings.keys():
            phase = self.phases()[obj.settings['phase']]
            return phase
        # Otherwise find it using bottom-up approach (i.e. look in phase keys)
        for phase in self.phases().values():
            if ('pore.'+obj.name in phase) or ('throat.'+obj.name in phase):
                return phase
        # If all else fails, throw an exception
        raise Exception('Cannot find a phase associated with '+obj.name)

    def find_geometry(self, physics):
        r"""
        Find the Geometry associated with a given Physics

        Parameters
        ----------
        physics : OpenPNM Physics Object
            Must be a Physics object

        Returns
        -------
        An OpenPNM Geometry object

        Raises
        ------
        If no Geometry object can be found, then an Exception is raised.

        """
        # If geometry happens to be in settings, look it up directly
        if 'geometry' in physics.settings.keys():
            geom = self.geometries()[physics.settings['geometry']]
            return geom
        # Otherwise, use the bottom-up approach
        for geo in self.geometries().values():
            if physics in self.find_physics(geometry=geo):
                return geo
        # If all else fails, throw an exception
        raise Exception('Cannot find a geometry associated with '+physics.name)

    def find_physics(self, geometry=None, phase=None):
        r"""
        Find the Physics object(s) associated with a given Geometry, Phase,
        or combination.

        Parameters
        ----------
        geometry : OpenPNM Geometry Object
            The Geometry object for which the Physics object(s) are sought

        phase : OpenPNM Phase Object
            The Phase object for which the Physics object(s) are sought

        Returns
        -------
        A list containing the Physics object(s).  If only a ``geometry`` is
        specified the the Physics for all Phases is returned.  If only a
        ``phase`` is specified, then the Physics for all Geometries is
        returned.  If both ``geometry`` and ``phase`` is specified then
        the list only contains a single Physics.  If no Physics is found, the
        the list will be empty.  See the Notes section for more information.

        See Also
        --------
        grid

        Notes
        -----
        The Project has an ``grid`` attribute that shows the association of
        all objects.  If each Geometry represents a row and each Phase is a
        column, then each row/col intersection represents a Physics. This
        method finds the PHysics' at each intersection

        """

        if geometry and phase:
            physics = self.find_physics(geometry=geometry)
            phases = list(self.phases().values())
            phys = physics[phases.index(phase)]
            return phys
        elif geometry:
            result = []
            net = self.network
            geoPs = net['pore.'+geometry.name]
            geoTs = net['throat.'+geometry.name]
            for phase in self.phases().values():
                physics = self.find_physics(phase=phase)
                temp = None
                for phys in physics:
                    Ps = phase.map_pores(pores=phys.Ps, origin=phys)
                    physPs = phase.tomask(pores=Ps)
                    Ts = phase.map_throats(throats=phys.Ts, origin=phys)
                    physTs = phase.tomask(throats=Ts)
                    if np.all(geoPs == physPs) and np.all(geoTs == physTs):
                        temp = phys
                result.append(temp)
            return result
        elif phase:
            names = set(self.physics().keys())
            keys = set([item.split('.')[-1] for item in phase.keys()])
            hits = names.intersection(keys)
            phys = [self.physics().get(i, None) for i in hits]
            return phys
        else:
            phys = list(self.physics().values())
            return phys

    def _validate_name(self, name):
        if name in self.names:
            raise Exception('An object already exists named '+name)
        for item in self:
            for key in item.keys():
                if key.split('.')[1] == name:
                    raise Exception('A property/label is already named '+name)

    def _generate_name(self, obj):
        prefix = obj.settings['prefix']
        num = str(len([item for item in self if item._isa() == obj._isa()]))
        name = prefix + '_' + num.zfill(2)
        return name

    @property
    def names(self):
        names = [i.name for i in self]
        return names

    def purge_object(self, obj, deep=False):
        r"""
        Remove an object from the Project.  This removes all references to
        the object from all other objects (i.e. removes labels)

        Parameters
        ----------
        obj : OpenPNM Object
            The object to purge

        deep : boolean
            A flag that indicates whether to remove associated objects.
            If ``True``, then removing a Geometry or Phase also removes
            the associated Physics objects.  If ``False`` (default) then
            only the given object is removed, along with its labels in all
            associated objects.  Removing a Physics always keeps associated
            Geometry and Phases since they might also be associated with other
            Physics objects.

        Raises
        ------
        An Exception is raised if the object is a Network.

        Notes
        -----
        For a clearer picture of this logic, type ``print(project.grid)`` at
        the console.  A deep purge of a Geometry is like removing a row, while
        a Phase is like removing a column.

        """
        if obj._isa() in ['physics', 'algorithm']:
            self._purge(obj)
        if obj._isa() == 'geometry':
            if deep:
                physics = self.find_physics(geometry=obj)
                for phys in physics:
                    self._purge(self.physics()[phys.name])
            self._purge(obj)
        if obj._isa() == 'phase':
            if deep:
                physics = self.find_physics(phase=obj)
                for phys in physics:
                    self._purge(self.physics()[phys.name])
            self._purge(obj)
        if obj._isa() == 'network':
            raise Exception('Cannot purge a network, just make a new project')

    def _purge(self, obj):
        for item in self:
            for key in list(item.keys()):
                if key.split('.')[-1] == obj.name:
                    del item[key]
        super().remove(obj)

    def save_object(self, obj):
        r"""
        Saves the given object to a file

        Parameters
        ----------
        obj : OpenPNM object
            The file to be saved.  Depending on the object type, the file
            extension will be one of 'net', 'geo', 'phase', 'phys' or 'alg'.
        """
        if not isinstance(obj, list):
            obj = [obj]
        for item in obj:
            filename = item.name + '.' + item.settings['prefix']
            with open(filename, 'wb') as f:
                pickle.dump({item.name: item}, f)

    def load_object(self, filename):
        r"""
        Loads a single object from a file

        Parameters
        ----------

        """
        with open(filename, 'rb') as f:
            d = pickle.load(f)
        for item in d.keys():
            self.extend(d[item])

    def save_project(self, filename=''):
        r"""
        Save the current project to a ``pnm`` file.

        Parameters
        ----------
        filename : string or path object
            The name of the file.  Can include an absolute or relative path
            as well.  If only a filename is given it will be saved in the
            current working directory.

        """
        ws.save_project(project=self, filename=filename)

    def _new_object(self, objtype, name=None):
        r"""
        """
        if objtype.startswith('net'):
            obj = openpnm.network.GenericNetwork(project=self, name=name)
        elif objtype.startswith('geo'):
            obj = openpnm.geometry.GenericGeometry(project=self, name=name)
        elif objtype.startswith('pha'):
            obj = openpnm.phases.GenericPhase(project=self, name=name)
        elif objtype.startswith('phy'):
            obj = openpnm.physics.GenericPhysics(project=self, name=name)
        elif objtype.startswith('alg'):
            obj = openpnm.algorithm.GenericAlgorithm(project=self, name=name)
        else:
            obj = openpnm.core.Base(project=self, name=name)
        return obj

    def import_data(self, filename):
        r"""
        """
        raise NotImplementedError('Use the io module to import data')

    def export_data(self, network=None, phases=[], filename=None,
                    filetype='vtp'):
        r"""
        Export the pore and throat data from the given object(s) into the
        specified file and format.

        Parameters
        ----------
        network: OpenPNM Network Object
            The network containing the data to be stored

        phases : list of OpenPNM Phase Objects
            The data on each supplied phase will be added to file

        filename : string
            The file name to use.  If none is supplied then one will be
            automatically generated based on the name of the project
            containing the supplied Network, with the date and time appended.

        filetype : string
            Which file format to store the data.  If a valid extension is
            included in the ``filename``, this is ignored.  Option are:

            **'vtk'** : (default) The Visualization Toolkit format, used by
            various softwares such as Paraview.  This actually produces a 'vtp'
            file.  NOTE: This can be quite slow since all the data is written
            to a simple text file.  For large data simulations consider
            'xdmf'.

            **'csv'** : The comma-separated values format, which is easily
            openned in any spreadsheet program.  The column names represent
            the property name, including the type and name of the object to
            which they belonged, all separated by the pipe character.

            **'xmf'** : The extensible data markup format, is a very efficient
            format for large data sets.  This actually results in the creation
            of two files, the *xmf* file and an associated *hdf* file.  The
            *xmf* file contains instructions for looking into the *hdf* file
            where the data is stored. Paraview opens the *xmf* format natively,
            and is very fast.

            **'mat'** : Matlab 'mat-file', which can be openned in Matlab.

        Notes
        -----
        This is a helper function for the actual functions in the IO module.
        For more control over the format of the output, and more information
        about the format refer to ``openpnm.io``.

        """
        project = network.project
        if filename is None:
            filename = project.name + '_' + time.strftime('%Y%b%d_%H%M%p')
        # Infer filetype from extension on file name...if given
        if '.' in filename:
            exts = ['vtk', 'vtp', 'vtu', 'csv', 'xmf', 'xdmf', 'hdf', 'hdf5',
                    'h5', 'mat']
            if filename.split('.')[-1] in exts:
                filename, filetype = filename.rsplit('.', 1)
        if filetype.lower() in ['vtk', 'vtp', 'vtu']:
            openpnm.io.VTK.save(network=network, phases=phases,
                                filename=filename)
        if filetype.lower() == 'csv':
            openpnm.io.CSV.save(network=network, phases=phases,
                                filename=filename)
        if filetype.lower() in ['xmf', 'xdmf']:
            openpnm.io.XDMF.save(network=network, phases=phases,
                                 filename=filename)
        if filetype.lower() in ['hdf5', 'hdf', 'h5']:
            f = openpnm.io.HDF5.to_hdf5(network=network, phases=phases,
                                        filename=filename)
            f.close()
        if filetype.lower() == 'mat':
            openpnm.io.MAT.save(network=network, phases=phases,
                                filename=filename)

    def _dump_data(self, mode=['props']):
        r"""
        Dump data from all objects in project to an HDF5 file.  Note that
        'pore.coords', 'throat.conns', 'pore.all', 'throat.all', and all
        labels pertaining to the linking of objects are kept.

        Parameters
        ----------
        mode : string or list of strings
            The type of data to be dumped to the HDF5 file.  Options are:

            **'props'** : Numerical data such as 'pore.diameter'.  The default
            is only 'props'.

            **'labels'** : Boolean data that are used as labels.  Since this
            is boolean data it does not consume large amounts of memory and
            probably does not need to be dumped.

        See Also
        --------
        _fetch_data

        Notes
        -----
        In principle, after data is fetched from an HDF5 file, it should
        physically stay there until it's called upon.  This let users manage
        the data as if it's in memory, even though it isn't.  This behavior
        has not been confirmed yet, which is why these functions are hidden.

        """
        with h5py.File(self.name + '.hdf5') as f:
            for obj in self:
                for key in list(obj.keys()):
                    tempname = obj.name + '|' + '_'.join(key.split('.'))
                    arr = obj[key]
                    if 'U' in str(obj[key][0].dtype):
                        pass
                    elif 'all' in key.split('.'):
                        pass
                    else:
                        f.create_dataset(name='/'+tempname, shape=arr.shape,
                                         dtype=arr.dtype, data=arr)
            for obj in self:
                obj.clear(mode=mode)

    def _fetch_data(self):
        r"""
        Retrieve data from an HDF5 file and place onto correct objects in the
        project

        See Also
        --------
        _dump_data

        Notes
        -----
        In principle, after data is fetched from and HDF5 file, it should
        physically stay there until it's called upon.  This let users manage
        the data as if it's in memory, even though it isn't.  This behavior
        has not been confirmed yet, which is why these functions are hidden.

        """
        with h5py.File(self.name + '.hdf5') as f:
            # Reload data into project
            for item in f.keys():
                obj_name, propname = item.split('|')
                propname = propname.split('_')
                propname = propname[0] + '.' + '_'.join(propname[1:])
                self[obj_name][propname] = f[item]

    @property
    def network(self):
        net = list(self._get_objects_by_type('network').values())
        if len(net) > 0:
            net = net[0]
        else:
            net = None
        return net

    def geometries(self, name=None):
        if name:
            return self._get_object_by_name(name)
        else:
            return self._get_objects_by_type('geometry')

    def phases(self, name=None):
        if name:
            return self._get_object_by_name(name)
        else:
            return self._get_objects_by_type('phase')

    def physics(self, name=None):
        if name:
            return self._get_object_by_name(name)
        else:
            return self._get_objects_by_type('physics')

    def algorithms(self, name=None):
        if name:
            return self._get_object_by_name(name)
        else:
            return self._get_objects_by_type('algorithm')

    def _get_object_by_name(self, name):
        for item in self:
            if item.name == name:
                return item
        raise Exception('An object named ' + name + ' was not found')

    def _get_objects_by_type(self, objtype):
        return {item.name: item for item in self if item._isa(objtype)}

    def _set_comments(self, string):
        if hasattr(self, '_comments') is False:
            self._comments = {}
        self._comments[time.strftime('%c')] = string

    def _get_comments(self):
        if hasattr(self, '_comments') is False:
            self._comments = {}
        for key in list(self._comments.keys()):
            print(key, ': ', self._comments[key])

    comments = property(fget=_get_comments, fset=_set_comments)

    def _get_grid(self):
        grid = {}
        row = {phase: '' for phase in self.phases().keys()}
        for geo in self.geometries().values():
            grid[geo.name] = row.copy()
            for phase in self.phases().values():
                phys = self.find_physics(phase=phase, geometry=geo)
                if phys is None:
                    phys = ''
                else:
                    phys = phys.name
                grid[geo.name][phase.name] = phys
        grid = ProjectGrid(self.network.name, grid)
        return grid

    grid = property(fget=_get_grid)

    def __str__(self):
        s = []
        hr = '―'*78
        s.append(hr)
        s.append(' {0:<15} '.format('Object Name') +
                 '{0:<65}'.format('Object ID'))
        s.append(hr)
        for item in self:
            s.append(' {0:<15} '.format(item.name) +
                     '{0:<65}'.format(item.__repr__()))
        s.append(hr)
        return '\n'.join(s)

    def check_geometry_health(self):
        r"""
        Perform a check to find pores with overlapping or undefined Geometries

        Returns
        -------
        A HealthDict
        """
        health = HealthDict()
        health['overlapping_pores'] = []
        health['undefined_pores'] = []
        health['overlapping_throats'] = []
        health['undefined_throats'] = []
        geoms = self.geometries().keys()
        if len(geoms):
            net = self.network
            Ptemp = np.zeros((net.Np,))
            Ttemp = np.zeros((net.Nt,))
            for item in geoms:
                Pind = net['pore.'+item]
                Tind = net['throat.'+item]
                Ptemp[Pind] = Ptemp[Pind] + 1
                Ttemp[Tind] = Ttemp[Tind] + 1
            health['overlapping_pores'] = np.where(Ptemp > 1)[0].tolist()
            health['undefined_pores'] = np.where(Ptemp == 0)[0].tolist()
            health['overlapping_throats'] = np.where(Ttemp > 1)[0].tolist()
            health['undefined_throats'] = np.where(Ttemp == 0)[0].tolist()
        return health

    def check_physics_health(self, phase):
        r"""
        Perform a check to find pores which have overlapping or missing Physics

        Parameters
        ----------
        phase : OpenPNM Phase object
            The Phase whose Physics should be checked

        Returns
        -------
        A HealthDict

        """
        health = HealthDict()
        health['overlapping_pores'] = []
        health['undefined_pores'] = []
        health['overlapping_throats'] = []
        health['undefined_throats'] = []
        geoms = self.geometries().keys()
        if len(geoms):
            phys = self.find_physics(phase=phase)
            if len(phys) == 0:
                raise Exception(str(len(geoms))+' geometries were found, but' +
                                ' no physics')
            if None in phys:
                raise Exception('Undefined physics found, check the grid')
            Ptemp = np.zeros((phase.Np,))
            Ttemp = np.zeros((phase.Nt,))
            for item in phys:
                    Pind = phase['pore.'+item.name]
                    Tind = phase['throat.'+item.name]
                    Ptemp[Pind] = Ptemp[Pind] + 1
                    Ttemp[Tind] = Ttemp[Tind] + 1
            health['overlapping_pores'] = np.where(Ptemp > 1)[0].tolist()
            health['undefined_pores'] = np.where(Ptemp == 0)[0].tolist()
            health['overlapping_throats'] = np.where(Ttemp > 1)[0].tolist()
            health['undefined_throats'] = np.where(Ttemp == 0)[0].tolist()
        return health

    def _regenerate_models(self, objs=[], propnames=[]):
        r"""
        Can be used to regenerate models across all objects in the project.

        Parameters
        ----------
        objs : list of OpenPNM objects
            Can be used to specify which specific objects to regenerate.  The
            default is to regenerate all objects.  If a subset of objects is
            given, this function ensure they are generated in a sensible order
            such as any phases are done before any physics objects.

        propnames : list of strings, or string
            The specific model to regenerate.  If none are given then ALL
            models on all objects are regenerated.  If a subset is given,
            then only object that have a corresponding model are regenerated,
            to avoid any problems.  This means that a single model can be
            given, without specifying the objects.

        """
        objs = list(objs)
        if objs == []:
            objs = self
        if type(propnames) is str:
            propnames = [propnames]
        # Sort objs in the correct order (geom, phase, phys)
        net = [i for i in objs if i is self.network]
        geoms = [i for i in objs if i in self.geometries().values()]
        phases = [i for i in objs if i in self.phases().values()]
        phys = [i for i in objs if i in self.physics().values()]
        objs = net + geoms + phases + phys
        for obj in objs:
            if len(propnames):
                for model in propnames:
                    if model in obj.models.keys():
                        obj.regenerate_models(propnames=model)
            else:
                obj.regenerate_models()


class Grid(dict):

    def __init__(self, name='', *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = name

    def index(self):
        return list(self.keys())

    def header(self):
        d = []
        for item in self.keys():
            d.extend([i for i in self[item].keys()])
        return list(set(d))

    def row(self, name):
        return list(self[name].values())

    def col(self, name):
        col = []
        for row in self.index():
            col.append(self[row][name])
        return col

    def __str__(self):
        s = []
        hr = '―'*(16*(len(self.header())+1))
        s.append(hr)
        fmt = ["| {"+str(i)+":^13} " for i in range(len(self.header()))]
        cols = [item for item in self.header()]
        s.append('| {0:^13}'.format(self.name) +
                 ''.join(fmt).format(*cols) + '|')
        s.append(hr)
        for row in self.index():
            ind = '| {0:^13}'.format(row)
            s.append(ind + ''.join(fmt).format(*list(self[row].values()))+'|')
            s.append(hr)
        return '\n'.join(s)


class ProjectGrid(Grid):
    r"""
    This is a subclass of Grid, which adds the ability to lookup by geometries
    and phases, as more specific versions of rows and cols
    """

    def geometries(self):
        return self.index()

    def phases(self):
        return self.header()



"""
===============================================================================
module Dispersion: Class for solving advection diffusion
===============================================================================
"""

import numpy as np
from scipy.sparse.csgraph import laplacian
from openpnm.algorithms import ReactiveTransport
from openpnm.utils import logging
logger = logging.getLogger(__name__)


class Dispersion(ReactiveTransport):
    r'''
    A subclass of GenericTransport to simulate advection diffusion.

    Examples
    --------
    >>>

    '''
    def __init__(self, settings={}, **kwargs):
        super().__init__(**kwargs)
        self.settings.update({'quantity': 'pore.concentration',
                              'hydraulic_conductance':
                              'throat.hydraulic_conductance',
                              'diffusive_conductance':
                              'throat.diffusive_conductance',
                              'pressure': 'pore.pressure'})
        self.settings.update(settings)

    def _build_A(self, force=False):
        r"""
        """
        network = self.project.network
        phase = self.project.phases()[self.settings['phase']]
        conns = network['throat.conns']

        P = phase[self.settings['pressure']]
        gh = phase[self.settings['hydraulic_conductance']]
        gd = phase[self.settings['diffusive_conductance']]
        gd = np.tile(gd, 2)

        Qij = -gh*np.diff(P[conns], axis=1).squeeze()
        Qij = np.append(Qij, -Qij)
        Peij = Qij/gd

        Peij[(Peij < 1e-10) & (Peij >= 0)] = 1e-10
        Peij[(Peij > -1e-10) & (Peij <= 0)] = -1e-10
        Qij = Peij*gd

        if force:
            self._pure_A = None
        if self._pure_A is None:
            w = -Qij / (1 - np.exp(Peij))
            A = network.create_adjacency_matrix(weights=w)
            A = laplacian(A)
            self._pure_A = A
        self.A = self._pure_A.copy()



import numpy as np
import scipy.stats as spts


def constant(target, value):
    r"""
    Places a constant value into the target object

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    value : scalar
        The numerical value to apply

    Notes
    -----
    This model is mostly useless and for testing purposes, but might be used
    to 'reset' an array back to a default value.
    """
    return value


def product(target, prop1, prop2, **kwargs):
    r"""
    Calculates the product of multiple property values

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    prop1 : string
        The name of the first argument

    prop2 : string
        The name of the second argument

    Notes
    -----
    Additional properties can be specified beyond just ``prop1`` and ``prop2``
    by including additional arguments in the function call (i.e. ``prop3 =
    'pore.foo'``).
    """
    value = target[prop1]*target[prop2]
    for item in kwargs.values():
        value *= target[item]
    return value


def random(target, element, seed=None, num_range=[0, 1]):
    r"""
    Create an array of random numbers of a specified size.

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    seed : int
        The starting seed value to send to Scipy's random number generator.
        The default is None, which means different distribution is returned
        each time the model is run.

    num_range : list
        A two element list indicating the low and high end of the returned
        numbers.

    """
    range_size = num_range[1] - num_range[0]
    range_min = num_range[0]
    if seed is not None:
        np.random.seed(seed)
    value = np.random.rand(target._count(element),)
    value = value*range_size + range_min
    return value


def scaled(target, prop, factor):
    r"""
    Scales an existing value by a factor.  Useful for constricting some throat
    property.

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    prop : string
        The dictionary key of the array containing the values to be scaled.

    factor : scalar
        The factor by which the values should be scaled.
    """
    value = target[prop]*factor
    return value


def linear(target, m, b, prop):
    r"""
    Calculates a property as a linear function of a given property

    Parameters
    ----------
    target : OpenPNM Object
        The object for which these values are being calculated.  This
        controls the length of the calculated array, and also provides
        access to other necessary thermofluid properties.

    m, b : floats
        Slope and intercept of the linear corelation

    prop : string
        The dictionary key containing the independent variable or phase
        property to be used in the correlation.

    """
    x = target[prop]
    value = m*x + b
    return value


def polynomial(target, a, prop, **kwargs):
    r"""
    Calculates a property as a polynomial function of a given property

    Parameters
    ----------
    target : OpenPNM Object
        The object for which these values are being calculated.  This
        controls the length of the calculated array, and also provides
        access to other necessary thermofluid properties.

    a : array_like
        A list containing the polynomial coefficients, where element 0 in the
        list corresponds to a0 and so on.  Note that no entries can be skipped
        so 0 coefficients must be sent as 0.

    prop : string
        The dictionary key containing the independent variable or phase
        property to be used in the polynomial.

    """
    x = target[prop]
    value = 0.0
    for i in range(0, len(a)):
        value += a[i]*x**i
    return value


def weibull(target, seeds, shape, scale, loc):
    r"""
    Produces values from a Weibull distribution given a set of random numbers.

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    seeds : string, optional
        The dictionary key on the Geometry object containing random seed values
        (between 0 and 1) to use in the statistical distribution.

    shape : float
        This controls the skewness of the distribution, with 'shape' < 1 giving
        values clustered on the low end of the range with a long tail, and
        'shape' > 1 giving a more symmetrical distribution.

    scale : float
        This controls the width of the distribution with most of values falling
        below this number.

    loc : float
        Applies an offset to the distribution such that the smallest values are
        above this number.

    Examples
    --------
    The following code illustrates the inner workings of this function,
    which uses the 'weibull_min' method of the scipy.stats module.  This can
    be used to find suitable values of 'shape', 'scale'` and 'loc'.  Note that
    'shape' is represented by 'c' in the actual function call.

    >>> import scipy
    >>> func = scipy.stats.weibull_min(c=1.5, scale=0.0001, loc=0)
    >>> import matplotlib.pyplot as plt
    >>> fig = plt.hist(func.ppf(q=scipy.rand(10000)), bins=50)

    """
    seeds = target[seeds]
    value = spts.weibull_min.ppf(q=seeds, c=shape, scale=scale, loc=loc)
    return value


def normal(target, seeds, scale, loc):
    r"""
    Produces values from a Weibull distribution given a set of random numbers.

    Parameters
    ----------
    target : OpenPNM Object
        The object with which this function as associated.  This argument
        is required to (1) set number of values to generate (geom.Np or
        geom.Nt) and (2) provide access to other necessary values
        (i.e. geom['pore.seed']).

    seeds : string, optional
        The dictionary key on the Geometry object containing random seed values
        (between 0 and 1) to use in the statistical distribution.

    scale : float
        The standard deviation of the Normal distribution

    loc : float
        The mean of the Normal distribution

    Examples
    --------
    The following code illustrates the inner workings of this function,
    which uses the 'norm' method of the scipy.stats module.  This can
    be used to find suitable values of 'scale' and 'loc'.

    >>> import scipy
    >>> func = scipy.stats.norm(scale=.0001, loc=0.001)
    >>> import matplotlib.pyplot as plt
    >>> fig = plt.hist(func.ppf(q=scipy.rand(10000)), bins=50)

    """
    seeds = target[seeds]
    value = spts.norm.ppf(q=seeds, scale=scale, loc=loc)
    return value


def generic(target, seeds, func):
    r"""
    Accepts an 'rv_frozen' object from the Scipy.stats submodule and returns
    values from the distribution for the given seeds using the ``ppf`` method.

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    seeds : string, optional
        The dictionary key on the Geometry object containing random seed values
        (between 0 and 1) to use in the statistical distribution.

    func : object
        An 'rv_frozen' object from the Scipy.stats library with all of the
        parameters pre-specified.

    Examples
    --------
    The following code illustrates the process of obtaining a 'frozen' Scipy
    stats object, and visualizes the corresponding distribution using a
    Matplotlib histogram:

    >>> import scipy
    >>> func = scipy.stats.weibull_min(c=2, scale=.0001, loc=0)
    >>> import matplotlib.pyplot as plt
    >>> fig = plt.hist(func.ppf(q=scipy.rand(1000)), bins=50)

    """
    seeds = target[seeds]
    value = func.ppf(seeds)
    return value


def from_neighbor_throats(target, throat_prop='throat.seed', mode='min'):
    r"""
    Adopt a value from the values found in neighboring throats

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    throat_prop : string
        The dictionary key of the array containing the throat property to be
        used in the calculation.  The default is 'throat.seed'.

    mode : string
        Controls how the pore property is calculated.  Options are 'min',
        'max' and 'mean'.
    """
    network = target.project.network
    Ps = target.pores()
    data = target[throat_prop]
    neighborTs = network.find_neighbor_throats(pores=Ps,
                                               flatten=False,
                                               mode='intersection')
    values = np.ones((np.shape(Ps)[0],))*np.nan
    if mode == 'min':
        for pore in Ps:
            values[pore] = np.amin(data[neighborTs[pore]])
    if mode == 'max':
        for pore in Ps:
            values[pore] = np.amax(data[neighborTs[pore]])
    if mode == 'mean':
        for pore in Ps:
            values[pore] = np.mean(data[neighborTs[pore]])
    return values


def from_neighbor_pores(target, pore_prop='pore.seed', mode='min'):
    r"""
    Adopt a value based on the values in neighboring pores

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    pore_prop : string
        The dictionary key to the array containing the pore property to be
        used in the calculation.  Default is 'pore.seed'.

    mode : string
        Controls how the throat property is calculated.  Options are 'min',
        'max' and 'mean'.

    """
    network = target.project.network
    throats = network.throats(target.name)
    P12 = network.find_connected_pores(throats)
    pvalues = network[pore_prop][P12]
    if mode == 'min':
        value = np.amin(pvalues, axis=1)
    if mode == 'max':
        value = np.amax(pvalues, axis=1)
    if mode == 'mean':
        value = np.mean(pvalues, axis=1)
    return value



from .misc import poisson_conductance


def series_resistors(target,
                     pore_thermal_conductivity='pore.thermal_conductivity',
                     throat_thermal_conductivity='throat.thermal_conductivity',
                     throat_equivalent_area='throat.equivalent_area',
                     throat_conduit_lengths='throat.conduit_lengths'):
    r"""
    Calculate the thermal conductance of conduits in network, where a
    conduit is ( 1/2 pore - full throat - 1/2 pore ) based on the areas

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    pore_thermal_conductivity : string
        Dictionary key of the pore thermal conductivity values

    throat_thermal_conductivity : string
        Dictionary key of the throat thermal conductivity values

    throat_equivalent_area : string
        Dictionary key of the throat equivalent area values

    throat_conduit_lengths : string
        Dictionary key of the throat conduit lengths

    Notes
    -----
    (1) This function requires that all the necessary phase properties already
    be calculated.

    (2) This function calculates the specified property for the *entire*
    network then extracts the values for the appropriate throats at the end.

    """
    return poisson_conductance(target=target,
                               pore_diffusivity=pore_thermal_conductivity,
                               throat_diffusivity=throat_thermal_conductivity,
                               throat_equivalent_area=throat_equivalent_area,
                               throat_conduit_lengths=throat_conduit_lengths)



from numpy import pi as _pi
import numpy as _np


def spherical_pores(target, throat_area='throat.area',
                    pore_diameter='pore.diameter',
                    throat_conduit_lengths='throat.conduit_lengths'):
    r"""
    Calculate equivalent cross-sectional area for a conduit consisting of two
    spherical pores and a constant cross-section throat. This area can be later
    used to calculate hydraulic or diffusive conductance of the conduit.

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    throat_area : string
        Dictionary key of the throat area values

    pore_diameter : string
        Dictionary key of the pore diameter values

    """
    network = target.project.network
    throats = network.map_throats(throats=target.Ts, origin=target)
    cn = network['throat.conns'][throats]
    d1 = network[pore_diameter][cn[:, 0]]
    d2 = network[pore_diameter][cn[:, 1]]
    L1 = target[throat_conduit_lengths+'.pore1']
    L2 = target[throat_conduit_lengths+'.pore2']
    return {'pore1': d1*L1*_pi / (2*_np.arctanh(2*L1/d1)),
            'throat': target[throat_area],
            'pore2': d2*L2*_pi / (2*_np.arctanh(2*L2/d2))}


def truncated_pyramid(target, throat_area='throat.area',
                      pore_diameter='pore.diameter'):
    r"""
    Calculate equivalent cross-sectional area for a conduit consisting of two
    truncated pyramid pores and a constant cross-section throat. This area can
    be later used to calculate hydraulic or diffusive conductance of the
    conduit.

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    throat_area : string
        Dictionary key of the throat area values

    pore_diameter : string
        Dictionary key of the pore diameter values

    """
    network = target.project.network
    throats = network.map_throats(throats=target.Ts, origin=target)
    cn = network['throat.conns'][throats]
    d1 = network[pore_diameter][cn[:, 0]]
    d2 = network[pore_diameter][cn[:, 1]]
    td = _np.sqrt(target[throat_area])
    return {'pore1': d1*td,
            'throat': target[throat_area],
            'pore2': d2*td}


def circular_pores(target, throat_area='throat.area',
                   pore_diameter='pore.diameter',
                   throat_conduit_lengths='throat.conduit_lengths'):
    r"""
    Calculate equivalent cross-sectional area for a conduit consisting of two
    spherical pores and a constant cross-section throat. This area can be later
    used to calculate hydraulic or diffusive conductance of the conduit.

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    throat_area : string
        Dictionary key of the throat area values

    pore_diameter : string
        Dictionary key of the pore diameter values

    """
    network = target.project.network
    throats = network.map_throats(throats=target.Ts, origin=target)
    cn = network['throat.conns'][throats]
    d1 = network[pore_diameter][cn[:, 0]]
    d2 = network[pore_diameter][cn[:, 1]]
    L1 = target[throat_conduit_lengths+'.pore1']
    L2 = target[throat_conduit_lengths+'.pore2']
    return {'pore1': 2*L1 / (_np.arctan(2*L1/_np.sqrt(d1**2 - 4*L1**2))),
            'throat': target[throat_area],
            'pore2': 2*L2 / (_np.arctan(2*L2/_np.sqrt(d2**2 - 4*L2**2)))}



# -*- coding: utf-8 -*-
"""
===============================================================================
Boundary -- Subclass of GenericGeometry for Boundary Pores
===============================================================================

"""

from openpnm.models import geometry as gm
from openpnm.models import misc as mm
from openpnm.geometry import GenericGeometry


class Boundary(GenericGeometry):
    r"""
    Boundary subclass of GenericGeometry.

    Parameters
    ----------
    network : OpenPNM Network object
        The Network to which the Geometry object should be associated

    pores : array_like
        The pores indice where the Geometry should be applied

    throat : array_like
        The throat indices where the Geometry should be applied

    shape: str
        Stick and Ball or Cube and Cuboid? ('spheres','cubes')

    name : string, optional
        A name to help identify the object, if not given one is automatically
        generated

    project : OpenPNM Project, optional
        Either a Network or a Project must be specified

    Examples
    --------
    >>> import openpnm
    >>> pn = openpnm.network.Cubic(shape=[3, 3, 3])
    >>> Ps_int = pn.pores(labels=['top', 'bottom'], mode='not')
    >>> Ps_boun = pn.pores(labels=['top', 'bottom'], mode='union')
    >>> Ts_int = pn.throats(labels=['internal'])
    >>> Ts_boun = pn.throats(labels=['internal'], mode='not')
    >>> geo = openpnm.geometry.GenericGeometry(network=pn,
    ...                                        pores=Ps_int, throats=Ts_int)
    >>> boun = openpnm.geometry.Boundary(network=pn, pores=Ps_boun,
    ...                                  throats=Ts_boun)

    """

    def __init__(self, shape='spheres', **kwargs):
        super().__init__(**kwargs)
        self['pore.diameter'] = 1e-12
        self.add_model(propname='throat.diameter',
                       model=mm.from_neighbor_pores,
                       pore_prop='pore.diameter',
                       mode='max')
        self['pore.volume'] = 1e-36
        self['pore.seed'] = 1.0
        self['throat.seed'] = 1.0
        self['throat.volume'] = 1e-36
        self.add_model(propname='throat.length',
                       model=gm.throat_length.straight)
        if shape == 'spheres':
            self.add_model(propname='throat.area',
                           model=gm.throat_area.cylinder)
            self.add_model(propname='throat.surface_area',
                           model=gm.throat_surface_area.cylinder)
        elif shape == 'cubes':
            self.add_model(propname='throat.area',
                           model=gm.throat_area.cuboid)
            self.add_model(propname='throat.surface_area',
                            model=gm.throat_surface_area.cuboid)



from openpnm.algorithms import TransientReactiveTransport, AdvectionDiffusion
from openpnm.utils import logging
logger = logging.getLogger(__name__)


class TransientAdvectionDiffusion(TransientReactiveTransport,
                                  AdvectionDiffusion):
    r"""
    A subclass of GenericTransport to perform steady and transient simulations
    of pure diffusion and advection diffusion problems.

    """

    def __init__(self, settings={}, **kwargs):
        super().__init__(**kwargs)
        # Apply any received settings to overwrite defaults
        self.settings.update(settings)



import scipy as sp
import numpy as np
import openpnm as op
from pathlib import Path
from openpnm.utils import flat_list, logging
logger = logging.getLogger(__name__)


class GenericIO():

    @classmethod
    def save(cls):
        raise NotImplementedError("The \'save\' method for this class " +
                                  "does not exist yet")

    @classmethod
    def load(cls):
        raise NotImplementedError("The \'load\' method for this class " +
                                  "does not exist yet")

    @classmethod
    def _convert_data(cls, project):
        # Convert arrays of 1's and/or 0's to booleans
        for obj in project:
            for item in obj.keys():
                ra = obj[item]
                if (ra.dtype == int) and (ra.max() <= 1) and (ra.min() >= 0):
                    obj.update({item: ra.astype(bool)})
        return project

    @classmethod
    def _update_network(cls, network, net):
        # Infer Np and Nt from length of given prop arrays in file
        for el in ['pore', 'throat']:
            N = [sp.shape(net[i])[0] for i in net.keys() if i.startswith(el)]
            if N:
                N = sp.array(N)
                if sp.all(N == N[0]):
                    if (network._count(el) == N[0]) \
                            or (network._count(el) == 0):
                        network.update({el+'.all': sp.ones((N[0],),
                                                           dtype=bool)})
                        net.pop(el+'.all', None)
                    else:
                        raise Exception('Length of '+el+' data in file' +
                                        ' does not match network')
                else:
                    raise Exception(el+' data in file have inconsistent' +
                                    ' lengths')
        # Add data on dummy net to actual network
        for item in net.keys():
            # Try to infer array types and change if necessary
            # Chcek for booleans disguised and 1's and 0's
            if (net[item].dtype is int) and (net[item].max() == 1):
                net[item] = net[item].astype(bool)
            # Write data to network object
            network.update({item: net[item]})

        network._gen_ids()

        return network

    @classmethod
    def _parse_filename(cls, filename, ext=''):
        p = Path(filename)
        p = p.resolve()
        # If extension not part of filename
        if p.suffix == '':
            try:
                p = p.with_suffix(ext)
            except ValueError:
                p = p.with_suffix('.'+ext)
        return p

    @classmethod
    def _parse_args(cls, network, phases):
        # Convert network to a list, even if empty
        if network is None:
            network = []
        else:
            network = flat_list(network)
        # Ensure phases is a list, even if empty
        phases = flat_list(phases)
        # Get handle to project object
        if len(network) == 0:
            if len(phases) == 0:
                raise Exception('Must specify one of network or phase')
            else:
                project = phases[0].project
        else:
            project = network[0].project
        return (project, network, phases)



from openpnm.algorithms import TransientReactiveTransport, FickianDiffusion
from openpnm.utils import logging
logger = logging.getLogger(__name__)


class TransientFickianDiffusion(TransientReactiveTransport, FickianDiffusion):
    r"""
    A subclass of GenericTransport to simulate diffusion.

    """

    def __init__(self, settings={}, **kwargs):
        super().__init__(**kwargs)
        self.settings.update({'molar_density': 'pore.molar_density'})
        # Apply any received settings to overwrite defaults
        self.settings.update(settings)
        # Define _coef as the molar density
        phase = self.project.phases()[self.settings['phase']]
        self._coef = phase[self.settings['molar_density']]



from openpnm.algorithms import TransientReactiveTransport, Dispersion
from openpnm.utils import logging
logger = logging.getLogger(__name__)


class TransientDispersion(TransientReactiveTransport, Dispersion):
    r"""

    """

    def __init__(self, settings={}, **kwargs):
        super().__init__(**kwargs)
        # Apply any received settings to overwrite defaults
        self.settings.update(settings)



from openpnm.algorithms import ReactiveTransport
from openpnm.utils import logging
logger = logging.getLogger(__name__)


class FourierConduction(ReactiveTransport):
    r"""
    A subclass of GenericLinearTransport to simulate heat conduction.  The 2
    main roles of this subclass are to set the default property names and to
    implement a method for calculating the effective conductivity of the
    network.

    """

    def __init__(self, settings={}, **kwargs):
        super().__init__(**kwargs)
        self.settings.update({'quantity': 'pore.temperature',
                              'conductance': 'throat.thermal_conductance'})
        self.settings.update(settings)

    def setup(self, phase=None, quantity='', conductance='', **kwargs):
        r"""
        This method takes several arguments that are essential to running the
        algorithm and adds them to the settings.

        Parameters
        ----------
        phase : OpenPNM Phase object
            The phase on which the algorithm is to be run.  If no value is
            given, the existing value is kept.

        quantity : string
            The name of the physical quantity to be calcualted.  If no value is
            given, the existing value is kept.  The default value is
            ``'pore.temperature'``.

        conductance : string
            The name of the pore-scale transport conductance values.  These
            are typically calculate by a model attached to a *Physics* object
            associated with the given *Phase*.  If no value is given, the
            existing value is kept.  The default value is
            ``'throat.thermal_conductance'``.

        Notes
        -----
        Any additional arguments are added to the ``settings`` dictionary of
        the object.

        """
        if phase:
            self.settings['phase'] = phase.name
        if quantity:
            self.settings['quantity'] = quantity
        if conductance:
            self.settings['conductance'] = conductance
        super().setup(**kwargs)

    def calc_effective_conductivity(self):
        r"""
        This calculates the effective thermal conductivity.
        """
        return self._calc_eff_prop()



from openpnm.utils import logging as _logging
from openpnm.models import misc as _misc
import numpy as _np
_logger = _logging.getLogger(__name__)


def weibull(target, shape, scale, loc, seeds='pore.seed'):
    return _misc.weibull(target=target, shape=shape, scale=scale, loc=loc,
                         seeds=seeds)


weibull.__doc__ = _misc.weibull.__doc__


def normal(target, scale, loc, seeds='pore.seed'):
    return _misc.normal(target=target, scale=scale, loc=loc,
                        seeds=seeds)


normal.__doc__ = _misc.normal.__doc__


def random(target, seed=None, num_range=[0, 1]):
    return _misc.random(target=target, element='pore', seed=seed,
                        num_range=num_range)


normal.__doc__ = _misc.normal.__doc__


def generic(target, func, seeds='pore.seed'):
    return _misc.generic(target=target, func=func, seeds=seeds)


generic.__doc__ = _misc.generic.__doc__


def largest_sphere(target, fixed_diameter='pore.fixed_diameter', iters=10):
    r"""
    Finds the maximum diameter pore that can be placed in each location without
    overlapping any neighbors.

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    fixed_diameter : string
        The dictionary key containing the pore diameter values already
        assigned to network, if any.

    iters : integer
        The number of iterations to perform when searching for maximum
        diameter.  This function iteratively grows pores until they touch
        their nearest neighbor, which is also growing, so this parameter limits
        the maximum number of iterations.  The default is 10, but 5 is usally
        enough.

    Notes
    -----
    This model looks into all pores in the network when finding the diameter.
    This means that when multiple Geometry objects are defined, it will
    consider the diameter of pores on adjacent Geometries. If no diameters
    have been assigned to these neighboring pores it will assume 0.  If
    diameter value are assigned to the neighboring pores AFTER this model is
    run, the pores will overlap.  This can be remedied by running this model
    again.

    """
    network = target.project.network
    try:
        # Fetch any existing pore diameters on the network
        D = network[fixed_diameter]
        # Set any unassigned values (nans) to 0
        D[_np.isnan(D)] = 0
    except KeyError:
        D = _np.zeros([network.Np, ], dtype=float)
    Ps = network.pores(target.name)
    C1 = network['pore.coords'][network['throat.conns'][:, 0]]
    C2 = network['pore.coords'][network['throat.conns'][:, 1]]
    L = _np.sqrt(_np.sum((C1 - C2)**2, axis=1))
    while iters >= 0:
        iters -= 1
        Lt = L - _np.sum(D[network['throat.conns']], axis=1)/2
        am = network.create_adjacency_matrix(weights=Lt, fmt='lil')
        D[Ps] = D[Ps] + _np.array([_np.amin(row) for row in am.data])[Ps]*0.95
    if _np.any(D < 0):
        _logger.warning('Negative pore diameters found!  Neighboring pores' +
                        ' are larger than the pore spacing.')
    return D[network.pores(target.name)]


def equivalent_diameter(target, pore_volume='pore.volume',
                        pore_shape='sphere'):
    r"""
    Calculates the diameter of a sphere or edge-length of a cube with same
    volume as the pore.

    Parameters
    ----------
    target : OpenPNM Geometry Object
        The Geometry object which this model is associated with. This controls
        the length of the calculated array, and also provides access to other
        necessary geometric properties.

    pore_volume : string
        The dictionary key containing the pore volume values

    pore_shape : string
        The shape of the pore body to assume when back-calculating from
        volume.  Options are 'sphere' (default) or 'cube'.

    """
    from scipy.special import cbrt
    pore_vols = target[pore_volume]
    if pore_shape.startswith('sph'):
        value = cbrt(6*pore_vols/_np.pi)
    elif pore_shape.startswith('cub'):
        value = cbrt(pore_vols)
    return value



import numpy as np
from openpnm.algorithms import GenericTransport
from openpnm.utils import logging
logger = logging.getLogger(__name__)


class ReactiveTransport(GenericTransport):
    r"""
    """

    def __init__(self, **kwargs):
        self.settings.update({'sources': [],
                              'tolerance': 0.001,
                              'max_iter': 10000,
                              'relaxation_source': 1,
                              'relaxation_quantity': 1})
        super().__init__(**kwargs)

    def set_source(self, propname, pores):
        r"""
        Applies a given source term to the specified pores

        Parameters
        ----------
        propname : string
            The property name of the source term model to be applied

        pores : array_like
            The pore indices where the source term should be applied

        Notes
        -----
        Source terms cannot be applied in pores where boundary conditions have
        already been set. Attempting to do so will result in an error being
        raised.

        """
        locs = self.tomask(pores=pores)

        if (not np.all(np.isnan(self['pore.bc_value'][locs]))) or \
           (not np.all(np.isnan(self['pore.bc_rate'][locs]))):
            raise Exception('Boundary conditions already present in given ' +
                            'pores, cannot also assign source terms')
        self[propname] = locs
        self.settings['sources'].append(propname)

    def _set_BC(self, pores, bctype, bcvalues=None, mode='merge'):
        # First check that given pores do not have source terms already set
        for item in self.settings['sources']:
            if np.any(self[item][pores]):
                raise Exception('Source term already present in given ' +
                                'pores, cannot also assign boundary ' +
                                'conditions')
        # Then call parent class function if above check passes
        super()._set_BC(pores=pores, bctype=bctype, bcvalues=bcvalues,
                        mode=mode)

    _set_BC.__doc__ = GenericTransport._set_BC.__doc__

    def _update_physics(self):
        phase = self.project.phases()[self.settings['phase']]
        physics = self.project.find_physics(phase=phase)
        for item in self.settings['sources']:
            # Regenerate models with new guess
            quantity = self.settings['quantity']
            # Put quantity on phase so physics finds it when regenerating
            phase[quantity] = self[quantity]
            # Regenerate models, on either phase or physics
            phase.regenerate_models(propnames=item)
            for phys in physics:
                phys.regenerate_models(propnames=item)

    def _apply_sources(self):
        if self.settings['t_scheme'] == 'cranknicolson':
            f1 = 0.5
        else:
            f1 = 1
        phase = self.project.phases()[self.settings['phase']]
        relax = self.settings['relaxation_source']
        for item in self.settings['sources']:
            Ps = self.pores(item)
            # Add S1 to diagonal of A
            # TODO: We need this to NOT overwrite the A and b, but create
            # copy, otherwise we have to regenerate A and b on each loop
            datadiag = self._A.diagonal().copy()
            # Source term relaxation
            S1_old = phase[item+'.'+'S1'][Ps].copy()
            S2_old = phase[item+'.'+'S2'][Ps].copy()
            self._update_physics()
            S1 = phase[item+'.'+'S1'][Ps]
            S2 = phase[item+'.'+'S2'][Ps]
            S1 = relax*S1 + (1-relax)*S1_old
            S2 = relax*S2 + (1-relax)*S2_old
            phase[item+'.'+'S1'][Ps] = S1
            phase[item+'.'+'S2'][Ps] = S2
            datadiag[Ps] = datadiag[Ps] - f1*S1
            # Add S1 to A
            self._A.setdiag(datadiag)
            # Add S2 to b
            self._b[Ps] = self._b[Ps] + f1*S2

    def run(self, x=None):
        r"""
        Builds the A and b matrices, and calls the solver specified in the
        ``settings`` attribute.

        Parameters
        ----------
        x : ND-array
            Initial guess of unknown variable

        """
        logger.info('―'*80)
        logger.info('Running ReactiveTransport')
        # Create S1 & S1 for 1st Picard's iteration
        if x is None:
            x = np.zeros(shape=[self.Np, ], dtype=float)
        self[self.settings['quantity']] = x
        self._update_physics()

        x = self._run_reactive(x=x)
        self[self.settings['quantity']] = x

    def _run_reactive(self, x):
        if x is None:
            x = np.zeros(shape=[self.Np, ], dtype=float)
        self[self.settings['quantity']] = x
        relax = self.settings['relaxation_quantity']
        res = 1e+06
        for itr in range(int(self.settings['max_iter'])):
            if res >= self.settings['tolerance']:
                logger.info('Tolerance not met: ' + str(res))
                self[self.settings['quantity']] = x
                self._build_A(force=True)
                self._build_b(force=True)
                self._apply_BCs()
                self._apply_sources()
                x_new = self._solve()
                # Relaxation
                x_new = relax*x_new + (1-relax)*self[self.settings['quantity']]
                self[self.settings['quantity']] = x_new
                res = np.sum(np.absolute(x**2 - x_new**2))
                x = x_new
            elif res < self.settings['tolerance']:
                logger.info('Solution converged: ' + str(res))
                break
        return x_new



import numpy as np
from scipy.spatial import ConvexHull
from transforms3d import _gohlketransforms as tr
from openpnm.utils import logging
import matplotlib.pyplot as plt
logger = logging.getLogger(__name__)


def PolyArea2D(pts):
    r"""
    returns the area of a 2D polygon given the set of points defining the
    convex hull in correct order
    Example
    ---------
    >>> import openpnm.utils.vertexops as vo
    >>> tri = np.array([[0, 0], [1, 2], [2, 0]])
    >>> vo.PolyArea2D(tri) == 2.0
    True
    """
    lines = np.hstack([pts, np.roll(pts, -1, axis=0)])
    area = 0.5*abs(sum(x1*y2-x2*y1 for x1, y1, x2, y2 in lines))
    return area


def PolyPerimeter2D(pts):
    r"""
    returns the perimeter of a 2D polygon given the set of points defining the
    convex hull in correct order
    Example
    ---------
    >>> import openpnm.utils.vertexops as vo
    >>> quad = np.array([[0, 0], [0, 1], [1, 1], [1, 0]])
    >>> vo.PolyPerimeter2D(quad) == 4.0
    True
    """
    lines = np.hstack([pts, np.roll(pts, -1, axis=0)])
    perimeter = sum(np.sqrt((x2-x1)**2+(y2-y1)**2) for x1, y1, x2, y2 in lines)
    return perimeter


def PolyWeightedCentroid2D(pts):
    r"""
    returns the centroid of a 2D polygon given the set of points defining the
    convex hull in correct order
    Example
    ---------
    >>> import openpnm.utils.vertexops as vo
    >>> quad = np.array([[0, 0], [0, 2], [2, 2], [2, 0]])
    >>> vo.PolyWeightedCentroid2D(quad) == [1.0,1.0]
    True
    """
    lines = np.hstack([pts, np.roll(pts, -1, axis=0)])
    twice_area = 0.0
    cx = 0.0
    cy = 0.0
    for x1, y1, x2, y2 in lines:
        f = x1*y2 - x2*y1
        cx += (x1+x2)*f
        cy += (y1+y2)*f
        twice_area += f
    A = 0.5*(twice_area)
    Cx = cx/(6*A)
    Cy = cy/(6*A)

    return [Cx, Cy]


def scale(network, scale_factor=[1, 1, 1], preserve_vol=False,
          linear_scaling=[False, False, False]):
    r"""
    A method for scaling the coordinates and vertices to create anisotropic
    networks.
    The original domain volume can be preserved by setting preserve_vol = True

    Example
    ---------
    >>> import openpnm as op
    >>> import openpnm.utils.vertexops as vo
    >>> import numpy as np
    >>> pn = op.network.Delaunay(num_points=100, shape=[3, 2, 1])
    >>> # op.topotools.label_faces(pn)
    >>> # B1 = pn.pores("left_boundary")
    >>> # B2 = pn.pores("right_boundary")
    >>> # Vol = vo.vertex_dimension(pn,B1,B2)
    >>> # vo.scale(network=pn,scale_factor=[2,1,1],preserve_vol=True)
    >>> # Vol2 = vo.vertex_dimension(pn,B1,B2)
    >>> # np.around(Vol-Vol2,5) == 0.0
    True
    >>> # vo.scale(network=pn,scale_factor=[2,1,1],preserve_vol=False)
    >>> # Vol3 = vo.vertex_dimension(pn,B1,B2)
    >>> # np.around(Vol3/Vol,5) == 2.0
    True

    """
    from scipy.special import cbrt
    import scipy as sp
    minmax = np.around(vertex_dimension(network=network,
                                        face1=network.pores(),
                                        parm='minmax'), 10)
    scale_factor = np.asarray(scale_factor)
    if preserve_vol is True:
        scale_factor = scale_factor/(cbrt(sp.prod(scale_factor)))

    lin_scale = _linear_scale_factor(network["pore.coords"], minmax,
                                     scale_factor, linear_scaling)

    network["pore.coords"] = network["pore.coords"]*lin_scale
    # Cycle through all vertices of all pores updating vertex values
    for pore in network.pores():
        for i, vert in network['pore.vert_index'][pore].items():
            vert_scale = _linear_scale_factor(vert, minmax,
                                              scale_factor, linear_scaling)
            network["pore.vert_index"][pore][i] = vert*vert_scale
    # Cycle through all vertices of all throats updating vertex values
    for throat in network.throats():
        for i, vert in network['throat.vert_index'][throat].items():

            vert_scale = _linear_scale_factor(vert, minmax,
                                              scale_factor, linear_scaling)
            network["throat.vert_index"][throat][i] = vert*vert_scale
    # Scale the vertices on the voronoi diagram stored on the network
    # These are used for adding boundaries on the Delaunay network class
    vert = network._vor.vertices
    vert_scale = _linear_scale_factor(vert, minmax, scale_factor,
                                      linear_scaling)
    network._vor.vertices = vert*vert_scale


def _linear_scale_factor(points=None,
                         minmax=[0, 1, 0, 1, 0, 1],
                         scale_factor=[1, 1, 1],
                         linear_scaling=[False, False, False]):
    r"""
    Work out the linear scale factor of a point or set of points based on the
    domain extent, an absolute scale factor and linear_scaling bools for each
    coordinate. If all False the absolute scaling is applied equally across the
    domain. If one linear_scaling boolean is True then a linear function is
    applied to scaling the co-ordinates along that axis. If more than one
    boolean is true then a combined linear function is applied
    """
    [xmin, xmax, ymin, ymax, zmin, zmax] = minmax
    max_array = np.array([(xmax-xmin), (ymax-ymin), (zmax-zmin)])
    pos_array = points/max_array
    shape = np.shape(points)
    if len(shape) == 1:
        combined_pos = np.ones([1])
        for i in range(3):
            if linear_scaling[i] is True:
                combined_pos *= pos_array[i]
        lin_scale = (scale_factor - 1)*combined_pos + 1
    else:
        combined_pos = np.ones([shape[0]])
        for i in range(3):
            if linear_scaling[i] is True:
                combined_pos *= pos_array[:, i]
        pos_array = np.vstack((combined_pos, combined_pos, combined_pos)).T
        lin_scale = (scale_factor - 1)*pos_array + 1

    return lin_scale


def vertex_dimension(network, face1=[], face2=[], parm='volume'):
    r"""
    Return the domain extent based on the vertices

    This function is better than using the pore coords as they may be far
    away from the original domain size.  And will alter the effective
    properties which should be based on the original domain sizes. Takes
    one or two sets of pores and works out different geometric properties
    if "length" is specified and two lists are given the planarity is
    determined and the appropriate length (x,y,z) is returned.  It should
    work the same as domain length and area if vertices are not in network
    by using coordinates.

    Example
    ----------
    >>> import openpnm as op
    >>> import openpnm.utils.vertexops as vo
    >>> pn = op.network.Delaunay(num_points=100, shape=[3, 2, 1])
    >>> # op.topotools.label_faces(pn)
    >>> # B1 = pn.pores('left_boundary')
    >>> # B2 = pn.pores('right_boundary')
    >>> # vo.vertex_dimension(pn, B1, B2,'volume')
    6.0
    >>> # vo.vertex_dimension(pn, B1, B2,'area')
    3.0
    >>> # vo.vertex_dimension(pn, B1, B2,'length')
    2.0
    >>> # vo.vertex_dimension(pn, B1, B2, 'area_xy')
    6.0
    >>> # vo.vertex_dimension(pn, B1, B2, 'area_yz')
    2.0
    >>> # vo.vertex_dimension(pn, B1, B2, 'area_xz')
    3.0
    >>> # vo.vertex_dimension(pn, B1, B2, 'minmax') == [0.0, 3.0, 0.0, 2.0, 0.0, 1.0]
    True
    """
    pores = np.array([], dtype=int)
    if 0 < len(face1):
        pores = np.hstack((pores, face1))
    if 0 < len(face2):
        pores = np.hstack((pores, face2))

    face1_coords = network['pore.coords'][face1]
    face2_coords = network['pore.coords'][face2]
    face1_planar = np.zeros(3)
    face2_planar = np.zeros(3)
    planar = np.zeros(3)
    for i in range(3):
        if len(np.unique(face1_coords[:, i])) == 1:
            face1_planar[i] = 1
        if len(np.unique(face2_coords[:, i])) == 1:
            face2_planar[i] = 1
    if 0 < len(face1) and 0 < len(face2):
        planar = face1_planar*face2_planar
    elif 0 < len(face1):
        planar = face1_planar
    elif 0 < len(face2):
        planar = face2_planar
    else:
        return 0

    if 'pore.vertices' in network.props():
        verts = []
        for pore in pores:
            for vert in np.asarray(list(network['pore.vertices'][pore])):
                verts.append(vert)
        verts = np.asarray(verts)
    else:
        verts = network['pore.coords'][pores]

    vx_min = verts[:, 0].min()
    vx_max = verts[:, 0].max()
    vy_min = verts[:, 1].min()
    vy_max = verts[:, 1].max()
    vz_min = verts[:, 2].min()
    vz_max = verts[:, 2].max()
    output = 0
    width = np.around(vx_max-vx_min, 10)
    depth = np.around(vy_max-vy_min, 10)
    height = np.around(vz_max-vz_min, 10)

    if parm == 'volume':
        output = width*depth*height
    elif parm == 'area_xy' or (parm == 'area' and planar[2] == 1):
        output = width*depth
    elif parm == 'area_xz' or (parm == 'area' and planar[1] == 1):
        output = width*height
    elif parm == 'area_yz' or (parm == 'area' and planar[0] == 1):
        output = depth*height
    elif parm == 'length_x' or (parm == 'length' and planar[0] == 1):
        output = width
    elif parm == 'length_y'or (parm == 'length' and planar[1] == 1):
        output = depth
    elif parm == 'length_z'or (parm == 'length' and planar[2] == 1):
        output = height
    elif parm == 'minmax':
        output = [vx_min, vx_max, vy_min, vy_max, vz_min, vz_max]

    return output


def porosity(network):
    r"""
    Return the porosity of the domain - sum of the pore volumes divided by
    domain volume
    """
    domain_vol = vertex_dimension(network, network.pores(), parm='volume')
    try:
        pore_vol = np.sum(network['pore.volume'])
    except KeyError:
        logger.error('Geometries must be assigned first')
        pore_vol = 0
    porosity = np.around(pore_vol/domain_vol, 3)
    return porosity


def pore2centroid(network):
    r"""
    Move the pore coordinate to the centroid of the pore vertices
    """
    for geom_name in network.geometries():
        geometry = network.geometries(geom_name)[0]
        if 'pore.centroid' in geometry.props():
            net_pores = network.pores(geometry.name)
            geom_pores = network.map_pors(poress=geometry.Ts, target=geometry)
            for i in range(len(geom_pores)):
                network['pore.coords'][net_pores[i]] = \
                    geometry['pore.centroid'][geom_pores[i]]


def tortuosity(network=None):
    r"""
    Calculate the tortuosity from the angle between throat vectors and
    principle axes
    """
    conns = network['throat.conns']
    va = network['throat.centroid'] - network['pore.centroid'][conns[:, 0]]
    vb = network['throat.centroid'] - network['pore.centroid'][conns[:, 1]]
    x = [1, 0, 0]
    y = [0, 1, 0]
    z = [0, 0, 1]
    f = 180 / np.pi
    theta_x_a = tr.angle_between_vectors(va, x, directed=False, axis=1)
    theta_x_b = tr.angle_between_vectors(vb, x, directed=False, axis=1)
    theta_x = (np.mean(theta_x_a[~np.isnan(theta_x_a)]) +
               np.mean(theta_x_b[~np.isnan(theta_x_b)])) / 2
    theta_y_a = tr.angle_between_vectors(va, y, directed=False, axis=1)
    theta_y_b = tr.angle_between_vectors(vb, y, directed=False, axis=1)
    theta_y = (np.mean(theta_y_a[~np.isnan(theta_y_a)]) +
               np.mean(theta_y_b[~np.isnan(theta_y_b)])) / 2
    theta_z_a = tr.angle_between_vectors(va, z, directed=False, axis=1)
    theta_z_b = tr.angle_between_vectors(vb, z, directed=False, axis=1)
    theta_z = (np.mean(theta_z_a[~np.isnan(theta_z_a)]) +
               np.mean(theta_z_b[~np.isnan(theta_z_b)])) / 2
    tot_angle = (theta_x+theta_y+theta_z)*f
    if 180 < tot_angle:
        logger.error('Something is wrong: ' + str(tot_angle))

    return 1 / np.cos(np.array([theta_x, theta_y, theta_z]))


def plot_throat(geometry, throats, fig=None):
    r"""
    Print a given throat or list of throats accepted as [1, 2, 3, ..., n]
    Original vertices plus offset vertices are rotated to align with
    the z-axis and then printed in 2D
    e.g vo.print_throat(geom, [34, 65, 99])
    """
    throat_list = []
    for throat in throats:
        if throat in range(geometry.num_throats()):
            throat_list.append(throat)
        else:
            logger.warn('Throat: ' + str(throat) + ' not part of geometry')
    if len(throat_list) > 0:
        verts = geometry['throat.vertices'][throat_list]
        offsets = geometry['throat.offset_vertices'][throat_list]
        normals = geometry['throat.normal'][throat_list]
        coms = geometry['throat.centroid'][throat_list]
        incentre = geometry['throat.incenter'][throat_list]
        inradius = 0.5*geometry['throat.indiameter'][throat_list]
        row_col = np.ceil(np.sqrt(len(throat_list)))
        for i in range(len(throat_list)):
            if fig is None:
                fig = plt.figure()
            ax = fig.add_subplot(row_col, row_col, i+1)
            vert_2D = rotate_and_chop(verts[i], normals[i], [0, 0, 1])
            hull = ConvexHull(vert_2D, qhull_options='QJ Pp')
            for simplex in hull.simplices:
                plt.plot(vert_2D[simplex, 0], vert_2D[simplex, 1], 'k-',
                         linewidth=2)
            plt.scatter(vert_2D[:, 0], vert_2D[:, 1])
            offset_2D = rotate_and_chop(offsets[i], normals[i], [0, 0, 1])
            offset_hull = ConvexHull(offset_2D, qhull_options='QJ Pp')
            for simplex in offset_hull.simplices:
                plt.plot(offset_2D[simplex, 0], offset_2D[simplex, 1],
                         'g-', linewidth=2)
            plt.scatter(offset_2D[:, 0], offset_2D[:, 1])
            # Make sure the plot looks nice by finding the greatest range of
            # points and setting the plot to look square
            xmax = vert_2D[:, 0].max()
            xmin = vert_2D[:, 0].min()
            ymax = vert_2D[:, 1].max()
            ymin = vert_2D[:, 1].min()
            x_range = xmax - xmin
            y_range = ymax - ymin
            if (x_range > y_range):
                my_range = x_range
            else:
                my_range = y_range
            lower_bound_x = xmin - my_range*0.5
            upper_bound_x = xmin + my_range*1.5
            lower_bound_y = ymin - my_range*0.5
            upper_bound_y = ymin + my_range*1.5
            plt.axis((lower_bound_x, upper_bound_x, lower_bound_y,
                      upper_bound_y))
            plt.grid(b=True, which='major', color='b', linestyle='-')
            centroid = rotate_and_chop(coms[i], normals[i], [0, 0, 1])
            incent = rotate_and_chop(incentre[i], normals[i], [0, 0, 1])
            plt.scatter(centroid[0][0], centroid[0][1])
            plt.scatter(incent[0][0], incent[0][1], c='r')
            # Plot incircle
            t = np.linspace(0, 2*np.pi, 200)
            u = inradius[i]*np.cos(t)+incent[0][0]
            v = inradius[i]*np.sin(t)+incent[0][1]
            plt.plot(u, v, 'r-')
            ax.ticklabel_format(style='sci', scilimits=(0, 0))

    else:
        logger.error("Please provide throat indices")
    return fig


def plot_pore(geometry, pores, fig=None, axis_bounds=None,
              include_points=False):
    r"""
    Print all throats around a given pore or list of pores accepted
    as [1, 2, 3, ..., n]
    e.g vo.print_pore(geom, [34, 65, 99])
    Original vertices plus offset vertices used to create faces and
    then printed in 3D
    To print all pores (n)
    pore_range = np.arange(0,n-1,1)
    vo.print_pore(geom, pore_range)
    """
    from mpl_toolkits.mplot3d.art3d import Poly3DCollection
    if len(pores) > 0:
        net = geometry.network
        net_pores = net.map_pores(pores=pores, origin=geometry)
        centroids = geometry['pore.centroid'][pores]
        coords = net['pore.coords'][net_pores]
        net_throats = net.find_neighbor_throats(pores=net_pores)
        throats = geometry.map_throats(throats=net_throats, origin=net)
        tcentroids = geometry["throat.centroid"][throats]
        # Can't create volume from one throat
        if 1 <= len(throats):
            verts = geometry['throat.vertices'][throats]
            normals = geometry['throat.normal'][throats]
            # Get verts in hull order
            ordered_verts = []
            for i in range(len(verts)):
                vert_2D = rotate_and_chop(verts[i], normals[i], [0, 0, 1])
                hull = ConvexHull(vert_2D, qhull_options='QJ Pp')
                ordered_verts.append(verts[i][hull.vertices])
            offsets = geometry['throat.offset_vertices'][throats]
            ordered_offs = []
            for i in range(len(offsets)):
                offs_2D = rotate_and_chop(offsets[i], normals[i], [0, 0, 1])
                offs_hull = ConvexHull(offs_2D, qhull_options='QJ Pp')
                ordered_offs.append(offsets[i][offs_hull.vertices])
            # Get domain extents for setting axis
            if axis_bounds is None:
                [xmin, xmax, ymin, ymax, zmin, zmax] = \
                    vertex_dimension(net, net_pores, parm='minmax')
            else:
                [xmin, xmax, ymin, ymax, zmin, zmax] = axis_bounds
            if fig is None:
                fig = plt.figure()
            ax = fig.gca(projection='3d')
            outer_items = Poly3DCollection(ordered_verts, linewidths=1,
                                           alpha=0.2, zsort='min')
            outer_face_colours = [(1, 0, 0, 0.01)]
            outer_items.set_facecolor(outer_face_colours)
            ax.add_collection(outer_items)
            inner_items = Poly3DCollection(ordered_offs, linewidths=1,
                                           alpha=0.2, zsort='min')
            inner_face_colours = [(0, 0, 1, 0.01)]
            inner_items.set_facecolor(inner_face_colours)
            ax.add_collection(inner_items)
            ax.set_xlim(xmin, xmax)
            ax.set_ylim(ymin, ymax)
            ax.set_zlim(zmin, zmax)
            if include_points:
                ax.scatter(centroids[:, 0],
                           centroids[:, 1],
                           centroids[:, 2],
                           c='y')
                ax.scatter(tcentroids[:, 0],
                           tcentroids[:, 1],
                           tcentroids[:, 2],
                           c='r')
                ax.scatter(coords[:, 0], coords[:, 1], coords[:, 2], c='b')
            ax.ticklabel_format(style='sci', scilimits=(0, 0))

        else:
            plot_throat(geometry, throats, fig)
    else:
        logger.error('Please provide pore indices')
    return fig


def rotate_and_chop(verts, normal, axis=[0, 0, 1]):
    r"""
    Method to rotate a set of vertices (or coords) to align with an axis
    points must be coplanar and normal must be given
    Chops axis coord to give vertices back in 2D
    Used to prepare verts for printing or calculating convex hull in order to
    arrange them in hull order for calculations and printing
    """
    xaxis = [1, 0, 0]
    yaxis = [0, 1, 0]
    zaxis = [0, 0, 1]
    angle = tr.angle_between_vectors(normal, axis)
    if angle == 0.0 or angle == np.pi:
        # We are already aligned
        facet = verts
    else:
        M = tr.rotation_matrix(tr.angle_between_vectors(normal, axis),
                               tr.vector_product(normal, axis))
        try:
            facet = np.dot(verts, M[:3, :3].T)
        except ValueError:
            pass

    try:
        x = facet[:, 0]
        y = facet[:, 1]
        z = facet[:, 2]
    except IndexError:
        x = facet[0]
        y = facet[1]
        z = facet[2]
    # Work out span of points and set axes scales to cover this and be
    # equal in both dimensions
    if axis == xaxis:
        output = np.column_stack((y, z))
    elif axis == yaxis:
        output = np.column_stack((x, z))
    elif axis == zaxis:
        output = np.column_stack((x, y))
    else:
        output = facet

    return output



r"""
**openpnm.algorithms**
----
The ``algorithms`` module contains classes for conducting transport simulations
on pore networks.
"""

from .GenericAlgorithm import GenericAlgorithm
from .GenericTransport import GenericTransport
from .ReactiveTransport import ReactiveTransport
from .TransientReactiveTransport import TransientReactiveTransport
from .StokesFlow import StokesFlow
from .FickianDiffusion import FickianDiffusion
from .TransientFickianDiffusion import TransientFickianDiffusion
from .AdvectionDiffusion import AdvectionDiffusion
from .TransientAdvectionDiffusion import TransientAdvectionDiffusion
from .Dispersion import Dispersion
from .TransientDispersion import TransientDispersion
from .FourierConduction import FourierConduction
from .OhmicConduction import OhmicConduction
from .OrdinaryPercolation import OrdinaryPercolation
from .InvasionPercolation import InvasionPercolation
from .MixedInvasionPercolation import MixedInvasionPercolation
from .Porosimetry import Porosimetry
